import unittest
import tempfile

import numpy as np
import h5py

from srlife import receiver

outer_radius = 1.0
thickness = 0.1
height = 100.0

nr = 10
nt = 10
nz = 10

inner_radius = outer_radius - thickness

period = 10.0
ndays = 10

nsteps = 10

times = np.linspace(0, period*ndays, ndays*nsteps)

panel_k = 0
tube_k = np.inf

def get_temp_hdf():
  return h5py.File(tempfile.mktemp(), "w")

def valid_pressure_bc():
  return receiver.PressureBC(times, np.zeros((len(times),)))

def valid_convective_bc(loc):
  if loc == "outer":
    r = outer_radius
  else:
    r = inner_radius
  return receiver.ConvectiveBC(r, height, nz, times, 
      np.zeros((len(times), nz)))

def valid_heatflux_bc(loc):
  if loc == "outer":
    r = outer_radius
  else:
    r = inner_radius
  return receiver.HeatFluxBC(r, height, nt, nz, times, 
      np.zeros((len(times), nt, nz)))

def valid_fixedtemp_bc(loc):
  if loc == "outer":
    r = outer_radius
  else:
    r = inner_radius
  return receiver.FixedTempBC(r, height, nt, nz, times, 
      np.zeros((len(times), nt, nz)))

def valid_tube(results = []):
  tube = receiver.Tube(outer_radius, thickness, height, nr, nt, nz)

  tube.set_times(times)

  for res in results:
    tube.add_results(res, np.zeros((len(times), nr, nt, nz)))

  tube.set_bc(valid_heatflux_bc("inner"), "inner")
  tube.set_bc(valid_convective_bc("outer"), "outer")
  tube.set_pressure_bc(valid_pressure_bc())

  return tube

def valid_tube_1D(results = []):
  tube = receiver.Tube(outer_radius, thickness, height, nr, nt, nz)

  tube.set_times(times)

  tube.make_1D(height/2, np.pi/3)

  for res in results:
    tube.add_results(res, np.zeros((len(times), nr)))

  return tube

def valid_panel(n = 0, results = []):
  panel = receiver.Panel(tube_k)

  for i in range(n):
    panel.add_tube(valid_tube(results))

  return panel

def valid_receiver(npanel = 0, ntube = 0, results = []):
  rec = receiver.Receiver(period, ndays, panel_k)

  for i in range(npanel):
    rec.add_panel(valid_panel(ntube, results))

  return rec

class TestReceiver(unittest.TestCase):
  def test_construct(self):
    rec = valid_receiver(10, 10, ['stress', 'strain'])

  def test_next_number(self):
    rec = valid_receiver(2)
    rec.add_panel(valid_panel(), name = "A")
    rec.add_panel(valid_panel())

    self.assertEqual(rec.npanels, 4)

    self.assertFalse("3" in rec.panels.keys())
    self.assertTrue("2" in rec.panels.keys())

  def test_store(self):
    base = valid_receiver(10,20,["stress", "strain"])
    f = get_temp_hdf()
    base.save(f)
    new = receiver.Receiver.load(f)

    self.assertTrue(base.close(new))

class TestPanel(unittest.TestCase):
  def test_construct(self):
    panel = valid_panel(2)

  def test_next_number(self):
    panel = valid_panel(2)
    panel.add_tube(valid_tube(), name = "A")
    panel.add_tube(valid_tube())

    self.assertEqual(panel.ntubes, 4)

    self.assertFalse("3" in panel.tubes.keys())
    self.assertTrue("2" in panel.tubes.keys())

  def test_store(self):
    base = valid_panel(5, ['stress'])
    
    f = get_temp_hdf()
    base.save(f)
    new = receiver.Panel.load(f)

    self.assertTrue(base.close(new))

class TestTube(unittest.TestCase):
  def test_construct(self):
    tube = valid_tube(["stress", "strain"])

  def test_invalid_size(self):
    tube = valid_tube(["stress"])
    with self.assertRaises(ValueError):
      tube.add_results("strain", np.zeros((1,)))

  def test_invalid_times(self):
    tube = valid_tube(["stress"])
    with self.assertRaises(ValueError):
      tube.set_times(np.linspace(0,1,10))

  def test_valid_bc(self):
    tube = valid_tube()
    tube.set_bc(valid_heatflux_bc("outer"), "outer")
    tube.set_bc(valid_heatflux_bc("inner"), "inner")

  def test_invalid_bc(self):
    tube = valid_tube()
    with self.assertRaises(ValueError):
      tube.set_bc(valid_heatflux_bc("outer"), "inner")

  def test_invalid_bc_location(self):
    tube = valid_tube()
    with self.assertRaises(ValueError):
      tube.set_bc(valid_heatflux_bc("outer"), "blah")

  def test_store(self):
    tube = valid_tube(['stress'])
    tube.set_bc(valid_heatflux_bc("outer"), "outer")
    
    f = get_temp_hdf()
    tube.save(f)
    new = receiver.Tube.load(f)

    self.assertTrue(tube.close(new))

  def test_store_1d(self):
    tube = valid_tube_1D(['stress'])
    tube.set_bc(valid_heatflux_bc("outer"), "outer")
    
    f = get_temp_hdf()
    tube.save(f)
    new = receiver.Tube.load(f)

    self.assertTrue(tube.close(new))

class TestPressureBC(unittest.TestCase):
  """
    Basic setup for PressureBC
  """
  def test_construct(self):
    bc = valid_pressure_bc()

  def test_store(self):
    f = get_temp_hdf()
    orig = valid_pressure_bc()
    orig.save(f)
    new = receiver.PressureBC.load(f)
    self.assertTrue(orig.close(new))

class TestHeatFluxBC(unittest.TestCase):
  """
    Test basic setup of a HeatFluxBC
  """
  def test_construct(self):
    bc = valid_heatflux_bc("outer")

  def test_wrong_size(self):
    with self.assertRaises(ValueError):
      bc = receiver.HeatFluxBC(inner_radius, height, nt, nz, times,
          np.zeros((1,)))

  def test_store(self):
    f = get_temp_hdf()
    orig = valid_heatflux_bc("outer")
    orig.save(f)
    new = receiver.ThermalBC.load(f)
    self.assertTrue(orig.close(new))

class TestFixedTempBC(unittest.TestCase):
  """
    Test basic setup of a FixedTempBC
  """
  def test_construct(self):
    bc = valid_fixedtemp_bc("outer")

  def test_wrong_size(self):
    with self.assertRaises(ValueError):
      bc = receiver.FixedTempBC(inner_radius, height, nt, nz, times,
          np.zeros((1,)))

  def test_store(self):
    f = get_temp_hdf()
    orig = valid_fixedtemp_bc("outer")
    orig.save(f)
    new = receiver.ThermalBC.load(f)
    self.assertTrue(orig.close(new))

class TestConvectiveBC(unittest.TestCase):
  """
    Test basic setup of a ConvectiveBC
  """
  def test_construct(self):
    bc = valid_convective_bc("outer")

  def test_wrong_size(self):
    with self.assertRaises(ValueError):
      bc = receiver.ConvectiveBC(inner_radius, height, nz, times,
          np.zeros((1,)))

  def test_store(self):
    f = get_temp_hdf()
    orig = valid_convective_bc("outer")
    orig.save(f)
    new = receiver.ThermalBC.load(f)
    self.assertTrue(orig.close(new))




import sys
sys.path.append('../../..')

import numpy as np
import matplotlib.pyplot as plt

from srlife import thermal, receiver, materials

class ManufacturedSolution:
  """
    A manufactured heat transport solution
  """
  def __init__(self, name, dim, solution, source):
    """
      Parameters:
        name:       descriptive name
        dim:        dimension of solution
        solution:   function of t, r, ...
        source      function of t, k, alpha, r, ...
        flux        function of t, k, alpha, r, ... (for BC)
    """
    self.name = name
    self.dim = dim
    self.soln = solution
    self.source = source

  def solve(self, solver, thermal, fluid, r = 1.0, t = 0.2, h = 1, time = 1, 
      ntime = 11, nr = 11, nt = 20, nz = 10, T0 = 0.0):
    """
      Generate the appropriate tube and solve with the provided solver
      
      Parameters:
        solver:         the thermal solver to test
        thermal:        the thermal model to test

      Other Parameters:
        r:              tube radius
        t:              tube thickness
        h:              tube height
        time:           maximum time
        ntime:          number of time steps
        nr:             number of radial increments
        nt:             number of circumferential increments
        nz:             number of axial increments
        T0:             initial temperature
    """
    tube = receiver.Tube(r, t, h, nr, nt, nz, T0)

    times = np.linspace(0, time, ntime)
    tube.set_times(times)

    if self.dim == 2:
      tube.make_2D(tube.h/2)
    elif self.dim == 1:
      tube.make_1D(tube.h/2, 0)

    def T0(*args):
      return self.soln(0, *args)

    def sfn(t, *args):
      k = thermal.conductivity(self.soln(t, *args))
      a = thermal.diffusivity(self.soln(t, *args))

      return self.source(t, k, a, *args)

    solver.solve(tube, thermal, materials.ConstantFluidMaterial({}), 
        source = sfn, T0 = T0, fix_edge = self.soln)

    return tube

  def plot_comparison(self, soln):
    """
      Plot a comparison to a manufactured solution

      Parameters:
        soln:           particular tube object with the solution
    """
    mesh = self._generate_mesh(soln.r, soln.t, soln.h, soln.times, 
        soln.nr, soln.nt, soln.nz)
    T = self.soln(*mesh)
    
    print(self.name)
    print("Max absolute error: %e" % np.max(np.abs(T - soln.results['temperature'])))
    keep = np.abs(T) > 0
    print("Max relative error: %e" % np.max(np.abs(T[keep] - soln.results['temperature'][keep])/np.abs(T[keep])))
    print("")

    plt.figure()
    if self.dim == 1:
      plot = [1, soln.nr // 2, -2]
      for p in plot:
        l, = plt.plot(mesh[0][:,p], soln.results['temperature'][:,p])
        plt.plot(mesh[0][:,p], T[:,p], ls = '--', color = l.get_color())
      plt.xlabel("Time")
      plt.ylabel("Temperature")
      plt.title("1D: left, middle, right")
    elif self.dim == 2:
      plot_r = [0,1, soln.nr // 2, -2,-1]
      plot_t = [0,1, soln.nt // 2, -2,-1]

      for rp in plot_r:
        for tp in plot_t:
          l, = plt.plot(mesh[0][:,rp,tp], soln.results['temperature'][:,rp,tp])
          plt.plot(mesh[0][:,rp,tp], T[:,rp, tp], ls = '--', color = l.get_color())
      plt.xlabel("Time")
      plt.ylabel("Temperature")
      plt.title("2D, r/theta slices")
    elif self.dim == 3:
      plot_r = [0,1, soln.nr // 2, -2,-1]
      plot_t = [0,1, soln.nt // 2, -2,-1]
      plot_z = [0,1, soln.nz // 2, -2,-1]
      for rp in plot_r:
        for tp in plot_t:
          for zp in plot_z:
            l, = plt.plot(mesh[0][:,rp,tp,zp], soln.results['temperature'][:,rp,tp,zp])
            plt.plot(mesh[0][:,rp,tp,zp], T[:,rp, tp,zp], ls = '--', color = l.get_color())
      plt.xlabel("Time")
      plt.ylabel("Temperature")
      plt.title("3D, r/theta/z slices")

  def assess_comparison(self, soln, tol, atol):
    """
      Return true if solution matches, false otherwise

      Parameters:
        soln:           particular tube object with the solution
        tol:            relative tolerance
    """
    mesh = self._generate_mesh(soln.r, soln.t, soln.h, soln.times, 
        soln.nr, soln.nt, soln.nz)
    T = self.soln(*mesh)

    return np.allclose(T, soln.results['temperature'], rtol = tol, atol = atol)

  def _generate_mesh(self, r, t, h, times, nr, nt, nz):
    """
      Generate the appropriate finite difference mesh for a particular problem
      
      Parameters:
        r:           radius
        t:           thickness
        h            height
        times:       discrete time steps
        nr:          number of radial increments
        nt:          number of circumferential increments
        nz:          number of axial increments
    """
    rs = np.linspace(r-t, r, nr)
    ts = np.linspace(0, 2*np.pi, nt + 1)[:-1]
    zs = np.linspace(0, h, nz)

    geom = [rs, ts, zs]

    return np.meshgrid(times, *geom[:self.dim], indexing = 'ij')

  def _generate_surface_mesh(self, t, h, times, nt, nz):
    """
      Generate the appropriate finite difference mesh for a particular problem
      
      Parameters:
        t:           thickness
        h:           height
        times:       discrete time steps
        nr:          number of radial increments
        nt:          number of circumferential increments
        nz:          number of axial increments
    """
    ts = np.linspace(0, 2*np.pi, nt + 1)[:-1]
    zs = np.linspace(0, h, nz)

    geom = [ts, zs]

    return np.meshgrid(times, *geom, indexing = 'ij')




import numpy as np

def differentiate(fn, x0, eps = 1.5e-6):
  if np.isscalar(x0):
    sx = (1,)
  else:
    sx = x0.shape

  f0 = fn(x0)

  if np.isscalar(f0):
    sf = (1,)
  else:
    sf = f0.shape

  tshape = sf + sx
  rshape = sx + sf

  D = np.zeros(rshape)

  for index in np.ndindex(sx):
    if np.isscalar(x0):
      diff = np.abs(x0 * eps)
    else:
      diff = np.abs(x0[index] * eps)
    if diff < eps:
      diff = eps
    
    if np.isscalar(x0):
      xp = diff
    else:
      xp = np.zeros(sx)
      xp[index] = diff

    fp = fn(xp+x0)

    D[index] = (fp - f0) / diff
  
  Df = np.zeros(tshape)
  # Reverse
  for ind1 in np.ndindex(sx):
    for ind2 in np.ndindex(sf):
      Df[ind2+ind1] = D[ind1+ind2]
  
  if np.isscalar(x0):
    return Df[0]
  else:
    return Df



"""
  This module defines 1D, 2D, and 3D thermal solvers.
"""

from abc import ABC, abstractmethod

import numpy as np
import numpy.linalg as la
import scipy.sparse as sp
import scipy.sparse.linalg as sla

from srlife import receiver, solverparams

class ThermalSolver(ABC):
  """
    When called this class takes as input:
      1) A Tube object
      2) A ThermalMaterial
      3) The FluidMaterial object 

    The solver must read the Tube abstraction parameter and 
    return a 3D, 2D, or 1D analysis, as appropriate
  """
  @abstractmethod
  def solve(self, tube, material, fluid):
    """
      Solve the thermal problem defined for a single tube

      Parameters:
        tube        the Tube object defining the geometry, loading,
                    and analysis abstraction
        material    the thermal material object describing the material
                    conductivity and diffusivity
        fluid       the fluid material object describing the convective
                    heat transfer coefficient
    """
    return

class FiniteDifferenceImplicitThermalSolver(ThermalSolver):
  """
    Solves the heat transfer problem using the finite difference method.

    Solver handles the *cylindrical* 1D, 2D, or 3D cases.
  """
  def __init__(self, pset = solverparams.ParameterSet(),
      rtol = 1.0e-6, atol = 1.0e-2, miter = 100, substep = 1, verbose = False):
    """
      Setup the solver

      Additional parameters:
        pset        object with solver parameters 
        rtol        iteration relative tolerance
        atol        iteration absolute tolerance
        miter       maximum iterations
        substep     divide user-provided time increments into smaller values 
        verbose     print a lot of debug info
    """
    self.rtol = pset.get_default("rtol", rtol)
    self.atol = pset.get_default("atol", atol)
    self.miter = pset.get_default("miter", miter)
    self.substep = pset.get_default("substep", substep)
    self.verbose = pset.get_default("verbose", verbose)

  def solve(self, tube, material, fluid, source = None, 
      T0 = None, fix_edge = None, rtol = 1e-6, atol = 1e-2, 
      miter = 100, substep = 1):
    """
      Solve the thermal problem defined for a single tube

      Parameters:
        tube        the Tube object defining the geometry, loading,
                    and analysis abstraction
        material    the thermal material object describing the material
                    conductivity and diffusivity
        fluid       the fluid material object describing the convective
                    heat transfer coefficient

      Other Parameters:
        source:     if present, the source term as a function of t and 
                    then the coordinates
        T0:         if present override the tube IC with a function of
                    the coordinates
        fix_edge:   an exact solution to fix edge BCs for testing
    """
    temperatures = FiniteDifferenceImplicitThermalProblem(tube, 
        material, fluid, source, T0, fix_edge, self.rtol, self.atol, 
        self.miter, self.substep, self.verbose).solve()

    tube.add_results("temperature", temperatures)

    return temperatures

class FiniteDifferenceImplicitThermalProblem:
  """
    The actual finite difference solver created to solve a single
    tube problem
  """
  def __init__(self, tube, material, fluid, source = None, T0 = None,
      fix_edge = None, rtol = 1.0e-6, atol = 1e-8, miter= 50, substep = 1,
      verbose = False):
    """
      Parameters:
        tube        Tube object to solve
        material    ThermalMaterial object
        fluid       FluidMaterial object

      Other Parameters:
        source      source function (t, r, ...)
        T0          initial condition function
        fix_edge:   an exact solution to fix edge BCs for testing
        rtol        relative tolerance
        atol        absolute tolerance
        miter       maximum iterations
        substep     divide user provided time increments into smaller steps
        verbose     print a lot of debug info
    """
    self.tube = tube
    self.material = material
    self.fluid = fluid
    
    self.rtol = rtol
    self.atol = atol
    self.miter = miter

    self.substep = substep

    self.verbose = verbose

    self.source_term = source
    self.T0 = T0
    self.fix_edge = fix_edge

    self.dr = self.tube.t / (self.tube.nr-1)
    self.dt = 2.0 * np.pi / (self.tube.nt)
    self.dz = self.tube.h / (self.tube.nz-1)

    self.dts = np.diff(self.tube.times)

    # Ghost
    self.dim = (self.tube.nr + 2, self.tube.nt+2, self.tube.nz + 2)

    if self.tube.abstraction == "3D":
      self.dim = self.dim
      self.nr, self.nt, self.nz = self.dim
      self.ndim = 3
      self.fdim = self.dim
    elif self.tube.abstraction == "2D":
      self.dim = self.dim[:2]
      self.nr, self.nt = self.dim
      self.nz = 1
      self.ndim = 2
      self.fdim = (self.nr, self.nt, 1)
    elif self.tube.abstraction == "1D":
      self.dim = self.dim[:1]
      self.nr, = self.dim
      self.nt = 1
      self.nz = 1
      self.ndim = 1
      self.fdim = (self.nr, 1, 1)
    else:
      raise ValueError("Thermal solver does not know how to handle"
      " abstraction %s" % self.tube.abstraction)

    # Useful for later
    self.mesh = self._generate_mesh()
    self.r = self.mesh[0].reshape(self.fdim)

    if self.ndim > 1:
      self.theta = self.mesh[1].reshape(self.fdim)
    else:
      self.theta = np.ones(self.r.shape) * self.tube.angle

    if self.ndim > 2:
      self.z = self.mesh[2].reshape(self.fdim)
    else:
      self.z = np.ones(self.r.shape) * self.tube.plane
   
  def _generate_mesh(self):
    """
      Produce the r, theta, z mesh
    """
    rs = np.linspace(self.tube.r - self.tube.t - self.dr, self.tube.r + self.dr, 
        self.tube.nr + 2)
    ts = np.linspace(-self.dt , 2.0*np.pi, self.tube.nt+2)
    zs = np.linspace(0 - self.dz, self.tube.h + self.dz, self.tube.nz + 2)

    geom = [rs, ts, zs]

    return np.meshgrid(*geom[:self.ndim], indexing = 'ij', copy = True)

  def solve(self):
    """
      Actually solve the problem...
    """
    # Setup the initial time
    T = np.zeros((self.tube.ntime,) + self.dim)
    if self.T0 is not None:
      T[0] = self.T0(*self.mesh)
    else:
      T[0] = self.tube.T0
    
    # Iterate through steps
    for i,(time,dt) in enumerate(zip(self.tube.times[1:],self.dts)):
      T[i+1] = self.solve_step_substep(T[i], time, dt)
    
    # Don't return ghost values
    if self.ndim == 3:
      return T[:,1:-1,1:-1,1:-1]
    elif self.ndim == 2:
      return T[:,1:-1,1:-1]
    else:
      return T[:,1:-1]

  def solve_step_substep(self, T_n, time, dt):
    """
      Do substepping, if requested by the user

      Parameters:
        T_n         previous full step temperature
        time        target time
        dt          target dt
    """
    T = np.copy(T_n)
    t_n = time - dt
    dti = dt / self.substep

    for i in range(1, self.substep + 1):
      t = t_n + dti * i
      T = self.solve_step(T, t, dti)

    return T

  def setup_step(self, T):
    """
      Setup common reusable arrays

      Parameters:
        T           temperature of interest
        time        current time
        dt          current dt
    """
    self.k = self.material.conductivity(T).reshape(self.fdim)
    self.a = self.material.diffusivity(T).reshape(self.fdim)

    # Useful matrix that gives you the actual dofs
    self.act = np.pad(np.ones(tuple(d-2 for d in self.dim)), [(1,1)] * self.ndim, 
        constant_values = [(0,0)] * self.ndim)

    self.act = self.act.reshape(self.fdim)

    self.ndof = T.size

  def _generate_A(self):
    """
      Generate the base differential operator for the current step
    """
    A = self.radial()

    if self.ndim > 1:
      A += self.circumfrential()
    
    if self.ndim > 2:
      A += self.axial()
    
    return A

  def _generate_id(self):
    """
      Return the non-boundary affecting ID matrix
    """
    return sp.diags([self.act.flatten()], offsets = (0,),
        shape = (self.ndof, self.ndof), format = 'coo')

  def _generate_source(self, time):
    """
      Generate the source part of the RHS

      Parameters:
        time        current time
    """
    if self.source_term:
      return (self.act * self.a / self.k * 
          self.source_term(time,*[self.r, self.theta, self.z][:self.ndim])).flatten()
    else:
      return np.zeros((self.ndof,))

  def _generate_prev_temp(self, T_n):
    """
      The RHS vector with the temperature contributions in the correct locations

      Parmaeters:
        T_n         previous temperatures
    """
    return (T_n * self.act).flatten()

  def _generate_bc_matrix(self, T_n, time, dt):
    """
      Generate the BC matrix terms (fixed)
    """
    M = self._ID_BC() + self._OD_BC()

    if self.ndim > 1:
      M += self._left_BC() + self._right_BC()

    if self.ndim > 2:
      M += self._top_BC() + self._bot_BC()

    return M

  def _generate_fixed_bc_RHS(self, T_n, time, dt):
    """
      Generate the constant (i.e. axial) contributions to the RHS

      Parameters:

      T_n       previous temperature
      time      current time
      dt        time increment
    """
    if self.ndim > 2:
      return self._top_BC_R(T_n, time, T_n) + self._bot_BC_R(T_n, time, T_n)
    else:
      return np.zeros((self.ndof,))
  
  # pylint: disable=too-many-locals
  def solve_step(self, T_n, time, dt):
    """
      Actually setup and solve for a single load step

      Parameters:
        T_n         previous temperatures
        time        current time
        dt          time increment
    """
    # Generic setup
    self.setup_step(T_n)

    # Add dimensions, if necessary
    T_n = T_n.reshape(self.fdim)

    # FD contributions
    A = self._generate_A()
    # Identity
    ID = self._generate_id()
    # Source term
    S = self._generate_source(time) 
    # Previous temp
    Tn = self._generate_prev_temp(T_n)
    
    # System without BCs
    M = (ID-A*dt)
    R = S*dt + Tn
    
    # Matrix and fixed RHS boundary contributions
    B = self._generate_bc_matrix(T_n, time, dt)
    M += B    
    BRF = self._generate_fixed_bc_RHS(T_n, time, dt)
    R += BRF
    
    # Dummy dofs
    D = self._generate_dummy_dofs()
    M += D

    # Covert over our fixed matrix
    M = M.tocsr()
    
    # This would be the iteration step
    T = np.copy(T_n)

    for i in range(self.miter):
      Ri = R + self._ID_BC_R(T, time, T_n) + self._OD_BC_R(T, time, T_n)
      J = self._d_ID_BC_R(T, time, T_n) + self._d_OD_BC_R(T, time, T_n)

      res = M.dot(T.flatten()) - Ri
      nr = la.norm(res)

      if i == 0:
        nr0 = nr

      if (nr < self.atol or nr/nr0 < self.rtol) and i > 0:
        break

      T -= sla.spsolve(M - J, res).reshape(self.fdim)
    else:
      raise RuntimeError("Too many iterations in newton solver!")
    
    return T.reshape(self.dim)
  
  # pylint: disable=too-many-branches
  def _generate_dummy_dofs(self):
    """
      Provide on-diagonal 1.0s for the dummy dofs
    """
    I = []
    J = []

    if self.ndim == 2:
      for i in self.dummy_loop_r():
        for j in self.dummy_loop_t():
          for k in self.dummy_loop_z():
            I.append(self.dof(i,j,k))
            J.append(self.dof(i,j,k))
    elif self.ndim == 3:
      for i in self.dummy_loop_r():
        for j in self.dummy_loop_t():
          for k in self.full_loop_z():
            I.append(self.dof(i,j,k))
            J.append(self.dof(i,j,k))
      for i in self.full_loop_r():
        for j in self.dummy_loop_t():
          for k in self.dummy_loop_z():
            I.append(self.dof(i,j,k))
            J.append(self.dof(i,j,k))
      for i in self.dummy_loop_r():
        for j in self.full_loop_t():
          for k in self.dummy_loop_z():
            I.append(self.dof(i,j,k))
            J.append(self.dof(i,j,k))

    return sp.coo_matrix((np.ones((len(I),)),(I,J)), shape = (self.ndof, self.ndof))

  def _ID_BC_R(self, T, time, T_n):
    """
      The inner diameter BC RHS contribution

      Parameters:
        T       current temperatures
        time    current time
        T_n     previous temperatures
    """
    R = np.zeros((self.ndof,))
    i = 0
    for j in self.loop_t():
      for k in self.loop_z():
        if self.fix_edge:
          R[self.dof(i,j,k)] = self.fix_edge(time, *[self.r[i,j,k], 
            self.theta[i,j,k], self.z[i,j,k]][:self.ndim])
        # Zero flux
        elif self.tube.inner_bc is None:
          R[self.dof(i,j,k)] = 0.0
        # Fixed temperature
        elif isinstance(self.tube.inner_bc, receiver.FixedTempBC):
          R[self.dof(i,j,k)] = self.tube.inner_bc.temperature(time, 
              self.theta[1,j,k], self.z[1,j,k])
        # Fixed flux
        elif isinstance(self.tube.inner_bc, receiver.HeatFluxBC):
          R[self.dof(i,j,k)] = -self.dr * self.tube.inner_bc.flux(time, 
              self.theta[1,j,k], self.z[1,j,k]) / self.k[1,j,k]
        # Convection
        elif isinstance(self.tube.inner_bc, receiver.ConvectiveBC):
          R[self.dof(i,j,k)] = self.dr * self.fluid.coefficient(self.material.name,
              T_n[1,j,k]) * (T[1,j,k] - self.tube.inner_bc.fluid_temperature(
                time, self.z[1,j,k])) / self.k[1,j,k]
        else:
          raise ValueError("Unknown boundary condition!")
    return R

  def _d_ID_BC_R(self, T, time, T_n):
    """
      Derivative of the inner diameter BC RHS contribution

      Parameters:
        T       current temperatures
        time    current time
        T_n     previous temperatures
    """
    I = []
    J = []
    D = []
    if isinstance(self.tube.inner_bc, receiver.ConvectiveBC):
      i = 0
      for j in self.loop_t():
        for k in self.loop_z():
          I.append(self.dof(i,j,k))
          J.append(self.dof(1,j,k))
          D.append(self.dr * self.fluid.coefficient(self.material.name, 
            T_n[1,j,k]) / self.k[1,j,k])
    
    return sp.coo_matrix((D,(I,J)), shape = (self.ndof, self.ndof))

  def _OD_BC_R(self, T, time, T_n):
    """
      The outer diameter BC RHS contribution

      Parameters:
        T       current temperatures
        time    current time
        T_n     previous temperatures
    """
    R = np.zeros((self.ndof,))
    i = self.nr-1
    for j in self.loop_t():
      for k in self.loop_z():
        if self.fix_edge:
          R[self.dof(i,j,k)] = self.fix_edge(time, *[self.r[i,j,k], 
            self.theta[i,j,k], self.z[i,j,k]][:self.ndim])
        # Zero flux
        elif self.tube.outer_bc is None:
          R[self.dof(i,j,k)] = 0.0
        # Fixed temperature
        elif isinstance(self.tube.outer_bc, receiver.FixedTempBC):
          R[self.dof(i,j,k)] = self.tube.outer_bc.temperature(time, 
              self.theta[self.nr-2,j,k], self.z[self.nr-2,j,k])
        # Fixed flux
        elif isinstance(self.tube.outer_bc, receiver.HeatFluxBC):
          R[self.dof(i,j,k)] = -self.dr * self.tube.outer_bc.flux(time, 
              self.theta[self.nr-2,j,k], self.z[self.nr-2,j,k]) / self.k[self.nr-2,j,k]
        # Convection
        elif isinstance(self.tube.outer_bc, receiver.ConvectiveBC):
          R[self.dof(i,j,k)] = (self.dr *
              self.fluid.coefficient(self.material.name, T_n[self.nr-2,j,k]) * 
              (T[self.nr-2,j,k] - 
                self.tube.outer_bc.fluid_temperature(time, self.z[self.nr-2,j,k])) / 
              self.k[self.nr-2,j,k])
        else:
          raise ValueError("Unknown boundary condition!")
    return R

  def _d_OD_BC_R(self, T, time, T_n):
    """
      Derivative of the outer diameter BC RHS contribution

      Parameters:
        T       current temperatures
        time    current time
        T_n     previous temperatures
    """
    I = []
    J = []
    D = []
    if isinstance(self.tube.outer_bc, receiver.ConvectiveBC):
      i = self.nr-1
      for j in self.loop_t():
        for k in self.loop_z():
          I.append(self.dof(i,j,k))
          J.append(self.dof(self.nr-2,j,k))
          D.append(self.dr * self.fluid.coefficient(self.material.name, 
            T_n[self.nr-2,j,k]) / self.k[self.nr-2,j,k])
    
    return sp.coo_matrix((D,(I,J)), shape = (self.ndof, self.ndof))

  def _ID_BC(self):
    """
      Inner diameter boundary condition contribution matrix
    """
    I = []
    J = []
    D = []
    i = 0
    for j in self.loop_t():
      for k in self.loop_z():
        if self.fix_edge:
          I.append(self.dof(i,j,k)) 
          J.append(self.dof(i,j,k))
          D.append(1.0)
        # Zero flux
        elif self.tube.inner_bc is None:
          I.append(self.dof(i,j,k)) 
          J.append(self.dof(1,j,k))
          D.append(1.0)
          I.append(self.dof(i,j,k)) 
          J.append(self.dof(0,j,k))
          D.append(-1.0)
        # Fixed temperature
        elif isinstance(self.tube.inner_bc, receiver.FixedTempBC):
          I.append(self.dof(i,j,k)) 
          J.append(self.dof(1,j,k))
          D.append(1.0)
        # Fixed flux
        elif isinstance(self.tube.inner_bc, receiver.HeatFluxBC):
          I.append(self.dof(i,j,k)) 
          J.append(self.dof(1,j,k))
          D.append(1.0)
          I.append(self.dof(i,j,k)) 
          J.append(self.dof(0,j,k))
          D.append(-1.0)
        # Convection
        elif isinstance(self.tube.inner_bc, receiver.ConvectiveBC):
          I.append(self.dof(i,j,k)) 
          J.append(self.dof(1,j,k))
          D.append(1.0)
          I.append(self.dof(i,j,k)) 
          J.append(self.dof(0,j,k))
          D.append(-1.0)
        else:
          raise ValueError("Unknown boundary condition!")

    return sp.coo_matrix((D,(I,J)), shape = (self.ndof, self.ndof))

  def _OD_BC(self):
    """
      Outer diameter contribution to the BC matrix
    """
    I = []
    J = []
    D = []
    i = self.nr-1
    for j in self.loop_t():
      for k in self.loop_z():
        if self.fix_edge:
          I.append(self.dof(i,j,k)) 
          J.append(self.dof(i,j,k))
          D.append(1.0)
        # Zero flux
        elif self.tube.outer_bc is None:
          I.append(self.dof(i,j,k)) 
          J.append(self.dof(self.nr-2,j,k))
          D.append(1.0)
          
          I.append(self.dof(i,j,k)) 
          J.append(self.dof(self.nr-1,j,k))
          D.append(-1.0)
        # Fixed temperature
        elif isinstance(self.tube.outer_bc, receiver.FixedTempBC):
          I.append(self.dof(i,j,k)) 
          J.append(self.dof(self.nr-2,j,k))
          D.append(1.0)
        # Fixed flux
        elif isinstance(self.tube.outer_bc, receiver.HeatFluxBC):
          I.append(self.dof(i,j,k)) 
          J.append(self.dof(self.nr-2,j,k))
          D.append(1.0)
          I.append(self.dof(i,j,k))
          J.append(self.dof(self.nr-1,j,k))
          D.append(-1.0)
        # Convection
        elif isinstance(self.tube.outer_bc, receiver.ConvectiveBC):
          I.append(self.dof(i,j,k)) 
          J.append(self.dof(self.nr-2,j,k))
          D.append(1.0)
          I.append(self.dof(i,j,k))
          J.append(self.dof(self.nr-1,j,k))
          D.append(-1.0)
        else:
          raise ValueError("Unknown boundary condition!")

    return sp.coo_matrix((D,(I,J)), shape = (self.ndof, self.ndof))

  def _left_BC(self):
    """
      Periodic contribution to the BC matrix
    """
    I = []
    J = []
    D = []
    j = 0
    for i in self.loop_r():
      for k in self.loop_z():
        I.append(self.dof(i,j,k))
        J.append(self.dof(i,j,k))
        D.append(1.0)

        I.append(self.dof(i,j,k))
        J.append(self.dof(i,self.nt-2,k))
        D.append(-1.0)

    return sp.coo_matrix((D,(I,J)), shape = (self.ndof, self.ndof))

  def _right_BC(self):
    """
      Periodic contribution to the BC matrix
    """
    I = []
    J = []
    D = []
    j = self.nt-1
    for i in self.loop_r():
      for k in self.loop_z():
        I.append(self.dof(i,j,k))
        J.append(self.dof(i,j,k))
        D.append(1.0)

        I.append(self.dof(i,j,k))
        J.append(self.dof(i,1,k))
        D.append(-1.0)

    return sp.coo_matrix((D,(I,J)), shape = (self.ndof, self.ndof))

  def _top_BC(self):
    """
      Axial top contribution to the BC matrix
    """
    I = []
    J = []
    D = []
    k = 0
    for i in self.loop_r():
      for j in self.loop_t():
        if self.fix_edge:
          I.append(self.dof(i,j,k))
          J.append(self.dof(i,j,k))
          D.append(1.0)
        else:
          I.append(self.dof(i,j,k))
          J.append(self.dof(i,j,1))
          D.append(1.0)
          I.append(self.dof(i,j,k))
          J.append(self.dof(i,j,0))
          D.append(-1.0)

    return sp.coo_matrix((D,(I,J)), shape = (self.ndof, self.ndof))

  def _bot_BC(self):
    """
      Axial bottom contribution to the BC matrix
    """
    I = []
    J = []
    D = []
    k = self.nz - 1
    for i in self.loop_r():
      for j in self.loop_t():
        if self.fix_edge:
          I.append(self.dof(i,j,k))
          J.append(self.dof(i,j,k))
          D.append(1.0)
        else:
          I.append(self.dof(i,j,k))
          J.append(self.dof(i,j,self.nz-2))
          D.append(1.0)
          I.append(self.dof(i,j,k))
          J.append(self.dof(i,j,self.nz-1))
          D.append(-1.0)

    return sp.coo_matrix((D,(I,J)), shape = (self.ndof, self.ndof))

  def _top_BC_R(self, T, time, T_n):
    """
      RHS contribution of the top axial BC

      Parameters:
        T           current temperatures
        time        current time
        T_n         previous temperatures
    """
    R = np.zeros((self.ndof,))
    k = 0
    for i in self.loop_r():
      for j in self.loop_t():
        if self.fix_edge:
          R[self.dof(i,j,k)] = self.fix_edge(time, 
              *[self.r[i,j,k], self.theta[i,j,k], 
                self.z[i,j,k]][:self.ndim])

    return R

  def _bot_BC_R(self, T, time, T_n):
    """
      RHS contribution of the bottom axial BC

      Parameters:
        T           current temperatures
        time        current time
        T_n         previous temperatures
    """
    R = np.zeros((self.ndof,))
    k = self.nz - 1
    for i in self.loop_r():
      for j in self.loop_t():
        if self.fix_edge:
          R[self.dof(i,j,k)] = self.fix_edge(time, 
              *[self.r[i,j,k], self.theta[i,j,k],
                self.z[i,j,k]][:self.ndim])

    return R

  def dof(self, i, j, k):
    """
      Return the DOF corresponding to the given grid position

      Parameters:
        i       r index
        j       theta index
        k       z index
    """
    return i * self.nt * self.nz + j * self.nz + k

  def loop_r(self):
    """
      Loop over non-ghost dofs
    """
    return range(1,self.nr-1)

  def loop_t(self):
    """
      Loop over non-ghost dofs
    """
    if self.ndim > 1:
      return range(1,self.nt-1)
    else:
      return [0]

  def loop_z(self):
    """
      Loop over non-ghost dofs
    """
    if self.ndim > 2:
      return range(1,self.nz-1)
    else:
      return [0]

  def full_loop_r(self):
    """
      Loop over all dofs
    """
    return range(0,self.nr)

  def full_loop_t(self):
    """
      Loop over all dofs
    """
    if self.ndim > 1:
      return range(0,self.nt)
    else:
      return [0]

  def full_loop_z(self):
    """
      Loop over all dofs
    """
    if self.ndim > 2:
      return range(0,self.nz)
    else:
      return [0]

  def dummy_loop_r(self):
    """
      Loop over ghost dofs
    """
    return [0,self.nr-1]

  def dummy_loop_t(self):
    """
      Loop over ghost dofs
    """
    return [0,self.nt-1]

  def dummy_loop_z(self):
    """
      Loop over ghost dofs
    """
    if self.ndim > 2:
      return [0,self.nz-1]
    else:
      return [0]

  def radial(self):
    """
      Insert the radial FD contribution into a sparse matrix
    """
    rh = (self.r[:-1] + self.r[1:]) / 2.0
    ah = (self.a[:-1] + self.a[1:]) / 2.0
    rhah = np.pad(rh * ah, ((1,1),(0,0),(0,0)), mode = 'edge')
   
    D1 = (self.act * rhah[:-1] / (self.r * self.dr**2.0)).flatten()[self.nt*self.nz:]
    D2 = -(self.act * (rhah[:-1] + rhah[1:]) / (self.r * self.dr**2.0)).flatten()
    D3 = (self.act * rhah[1:] / (self.r * self.dr**2.0)).flatten()[:-self.nt*self.nz]

    return sp.diags((D1,D2,D3), offsets = (-self.nt*self.nz,0,self.nt*self.nz), 
        shape = (self.ndof, self.ndof), format = 'coo')

  def circumfrential(self):
    """
      Insert the circumferential FD contribution into a coo matrix
    """
    ah = np.pad((self.a[:,:-1] + self.a[:,1:]) / 2.0,
        ((0,0),(1,1),(0,0)), mode = 'edge')

    D1 = (self.act * ah[:,:-1] / (self.r**2.0 * self.dt**2.0)).flatten()[self.nz:]
    D2 = -(self.act * (ah[:,:-1] + ah[:,1:]) / (self.r**2.0 * self.dt**2.0)).flatten()
    D3 = (self.act * ah[:,1:] / (self.r**2.0 * self.dt**2.0)).flatten()[:-self.nz]

    return sp.diags((D1,D2,D3), offsets = (-self.nz, 0, self.nz),
        shape = (self.ndof, self.ndof), format = 'coo')

  def axial(self):
    """
      Insert the axial FD contribution into a coo matrix
    """
    ah = np.pad((self.a[:,:,:-1] + self.a[:,:,1:]) / 2.0, 
        ((0,0),(0,0),(1,1)), mode = 'edge')

    D1 = (self.act * ah[:,:,:-1] / self.dz**2.0).flatten()[1:]
    D2 = (-self.act * (ah[:,:,:-1] + ah[:,:,1:]) / self.dz**2.0).flatten()
    D3 = (self.act * ah[:,:,1:] / self.dz**2.0).flatten()[:-1]

    return sp.diags([D1,D2,D3], offsets = (-1,0,1),
        shape = (self.ndof, self.ndof), format = 'coo')



#pylint: disable=unnecessary-lambda

"""
  A very small module defining a class to help organize solver parameters
"""

from collections import defaultdict

class ParameterSet(defaultdict):
  """ Recursive default dictionary that can be pickled

    Args:
      **kwargs: Arbitrary keyword arguments to be added to dictionary
  """
  def __init__(self, **kwargs):
    super().__init__(lambda: self.__class__(**kwargs))
    self.__dict__.update(kwargs)

  def __reduce__(self):
    """ Reduce to something pickle-able
    """
    return (type(self), (), None, None, iter(self.items()))

  def get_default(self, key, value):
    """ Get a key, if not present return the default value
    """
    if key in self.keys():
      return self[key]
    else:
      return value



#pylint: disable=unused-wildcard-import, wildcard-import

"""
  This module solves the receiver system or the single tube
  structural problem.
"""

from abc import ABC, abstractmethod

import copy

import scipy.sparse.linalg as spla
import numpy.linalg as la

import numpy as np

from skfem import *
from skfem import mesh, element
from skfem import helpers, utils

from neml import block

from srlife import solverparams

class TubeSolver(ABC):
  """
    This class takes as input:
      1) A state object containing whatever the solver needs to execute the solve
         These objects must contain a method for dumping the required results fields into
         a Tube object
      2) A input top displacement

    It must return:
      1) A copied and updated state
      2) The force on the top face
      3) The derivative of the force with respect to the displacement

      These three are grouped in a State object
  """
  @abstractmethod
  def solve(self, tube, i, state_n, dtop):
    """
      Solve the structural tube problem for a single time step

      Parameters:
        tube:       tube object with all bcs
        i:          time index to reference in tube results
        state:      state object
        dtop:       top displacement
    """
    return

  @abstractmethod
  def setup_tube(self, tube):
    """
      Setup all the quadrature results fields in the tube

      Parameters:
        tube:       tube object
    """
    return

  @abstractmethod
  def init_state(self, tube, mat):
    """
      Initialize the solver state

      Parameters:
        tube:       tube object
        mat:        NEML material
    """
    return

  @abstractmethod
  def dump_state(self, tube, i, state):
    """
      Update the required results fields in a tube object with the
      current state

      Parameters:
        tube:       tube to update
        i:          which time step this is
        state:      state object
    """
    return

def mesh_tube(tube):
  """
    Make a simple, regular Cartesian mesh of a Tube using:
      1D:   linear Lagrange (line)
      2D:   bilinear Lagrange (Quad4)
      3D:   trilinear Lagrange (Hex8)
    elements
  """
  if tube.abstraction == "3D":
    return mesh3D(tube)
  elif tube.abstraction == "2D":
    return mesh2D(tube)
  elif tube.abstraction == "1D":
    return mesh1D(tube)
  else:
    raise ValueError("Unknown tube abstraction %s!" % tube.abstraction)

def mesh1D(tube):
  """
    Make a 1D linear Lagrange mesh of a tube
  """
  return mesh.MeshLine(tube.mesh[0].flatten())

def mesh2D(tube):
  """
    Make a 2D Cartesian Lagrange mesh of a tube
  """
  rs = np.linspace(tube.r-tube.t, tube.r, tube.nr)
  ts = np.linspace(0, 2*np.pi, tube.nt+1)[:tube.nt]

  coords = np.swapaxes(np.array([[ri * np.cos(ts), ri * np.sin(ts)] 
    for ri in rs]), 0, 1).reshape(2, len(rs)*len(ts))
  
  conn = []
  for i in range(tube.nr-1):
    for j in range(tube.nt):
      conn.append(
          [i*tube.nt+j, i*tube.nt+((j+1)%tube.nt), 
           (i+1)*tube.nt+((j+1)%tube.nt), (i+1)*tube.nt+j])

  conn = np.array(conn, dtype = int).T

  return mesh.MeshQuad(coords, conn)

def mesh3D(tube):
  """
    Make a 3D Cartesian Lagrange mesh of a tube
  """
  rs = np.linspace(tube.r-tube.t, tube.r, tube.nr)
  ts = np.linspace(0, 2*np.pi, tube.nt+1)[:tube.nt] 
  zs = np.linspace(0, tube.h, tube.nz)
  
  npr = tube.nt * tube.nz
  npt = tube.nz

  coords = np.ascontiguousarray(
      np.array([[r*np.cos(t), r*np.sin(t), z] 
        for r in rs for t in ts for z in zs]).T)

  mapper = lambda r, c, h: r * npr + (c % tube.nt) * npt + h

  conn = []
  for i in range(tube.nr-1):
    for j in range(tube.nt):
      for k in range(tube.nz-1):
        conn.append(
            [
              mapper(i+1,j+1,k),
              mapper(i,j+1,k),
              mapper(i+1,j+1,k+1),
              mapper(i+1,j,k),
              mapper(i,j+1,k+1),
              mapper(i,j,k),
              mapper(i+1,j,k+1),
              mapper(i,j,k+1)
            ])

  conn = np.ascontiguousarray(np.array(conn, dtype = int).T)

  return mesh.MeshHex(coords, conn)

class PythonTubeSolver(TubeSolver):
  """
    Tube solver class coded up with scikit.fem and the scipy sparse solvers
  """
  def __init__(self, pset = solverparams.ParameterSet(), rtol = 1.0e-6,
      atol = 1.0e-8, qorder = 1, dof_tol = 1.0e-6, miter = 10, verbose = False):
    """
      Setup the solver with common parameters

      Additional Parameters:
        pset        parameter set with solver parameters
        rtol        relative tolerance for NR iterations
        atol        absolute tolerance for NR iterations
        qorder      quadrature order
        dof_tol     geometric tolerance on finding boundary
                    degrees of freedom
        miter       maximum newton-raphson iterations
        verbose     verbose solve
    """
    self.qorder = qorder
    self.rtol = pset.get_default("rtol", rtol)
    self.atol = pset.get_default("atol", atol)
    self.dof_tol = pset.get_default("dof_tol", dof_tol)
    self.miter = pset.get_default("miter", miter)
    self.verbose = pset.get_default("verbose", verbose)
    self.solver_options = {'rtol': self.rtol, 'atol': self.atol,
        'dof_tol': self.dof_tol, 'miter': self.miter, 'verbose': self.verbose}

  def setup_tube(self, tube):
    """
      Add all the quadrature result fields we will need

      Parameters:
        tube        tube object
    """
    suffixes = ['_xx', '_yy', '_zz', '_yz', '_xz', '_xy']
    fields = ['stress', 'strain', 'mechanical_strain', 'thermal_strain']

    for field in fields:
      for suffix in suffixes:
        tube.add_quadrature_results(field+suffix, np.zeros((tube.ntime,) + self.qshape(tube)))

    tube.add_quadrature_results("temperature", np.zeros((tube.ntime,) + self.qshape(tube)))

    suffixes = ["_x", "_y", "_z"]
    for i in range(tube.ndim):
      tube.add_results("disp"+suffixes[i], np.zeros((tube.ntime,) + tube.dim[:tube.ndim]))

  def qshape(self, tube):
    """
      Number of elements x number of quadrature points
    """
    if tube.dim[2] == 1:
      nz = 1
    else:
      nz = tube.dim[2] - 1
    nelem = (tube.dim[0] - 1) * (tube.dim[1]) * nz
    return (nelem, (self.qorder+1)**tube.ndim)

  def solve(self, tube, i, state_n, dtop):
    """
      Solve the structural tube problem for a single time step

      Parameters:
        tube:       tube object with all bcs
        i:          time index to reference in tube results
        state:      state object
        dtop:       top disoplacement
    """
    state_np1 = state_n.copy()

    t_np1 = tube.times[i]
    t_n = tube.times[i-1]

    if 'temperature' in tube.results:
      state_np1.temperature = self._res2quad(state_np1,
          self._tube2fea(tube, tube.results['temperature'][i]))
      state_n.temperature = self._res2quad(state_n,
          self._tube2fea(tube, tube.results['temperature'][i-1]))
    else:
      state_np1.temperature = np.zeros(state_np1.temperature.shape)
      state_n.temperature = np.zeros(state_n.temperature.shape)

    if tube.pressure_bc:
      p_np1 = tube.pressure_bc.pressure(t_np1)
      p_n = tube.pressure_bc.pressure(t_n)
    else:
      p_np1 = 0.0
      p_n = 0.0

    if tube.ndim == 1:
      solve_python_1d(state_n, t_n, p_n, state_np1, t_np1, p_np1, dtop,
          self.solver_options)
    elif tube.ndim == 2:
      solve_python_2d(state_n, t_n, p_n, state_np1, t_np1, p_np1, dtop,
          self.solver_options)
    elif tube.ndim == 3:
      solve_python_3d(state_n, t_n, p_n, state_np1, t_np1, p_np1, 
          dtop, self.solver_options)
    else:
      raise ValueError("Unknown dimension %i" % tube.ndim)

    return state_np1

  def init_state(self, tube, mat, i = None):
    """
      Initialize the solver state

      Parameters:
        tube:       tube object
        mat:        NEML material

      Additional Parameters:
        i:          if not none, also dump state into the tube
    """
    return PythonTubeSolver.State(tube, mat, self.qorder, i = i)

  def dump_state(self, tube, i, state):
    """
      Update the required results fields in a tube object with the
      current state

        Parameters:
        tube:       tube to update
        i:          which time step this is
        state:      state object
    """
    # Displacement
    suffixes = ["_x", "_y", "_z"]
    for k in range(state.ndim):
      tube.results["disp"+suffixes[k]][i] = self._fea2tube(tube,state.displacements[k::state.ndim])
    
    # The tensor fields
    order = ['_xx', '_yy', '_zz', '_yz', '_xz', '_xy']
    inds = [(0,0),(1,1),(2,2),(1,2),(0,2),(0,1)]
    fields = ['stress', 'strain', 'mechanical_strain', 'thermal_strain']
    data = [state.stress, state.strain, state.mechanical_strain, 
        state.thermal_strain]
    for f,d in zip(fields, data):
      for ind,o in zip(inds,order):
        tube.quadrature_results[f+o][i] = self._fea2tube_element(tube, d[ind])
    
    tube.quadrature_results["temperature"][i] = self._fea2tube_element(tube, state.temperature)

  def _fea2tube_element(self, tube, f):
    """
      Rearrange the elements (first index) of the field to match the vtk
      convention
    """
    return f

  def _tube2fea(self, tube, f):
    """
      Map a result field in the tube to the flat FEA vector
    """
    return f.flatten()

  def _fea2tube(self, tube, f):
    """
      Map a result field in the FEA to the right shape for the tube
    """
    return f.reshape(tube.dim[:tube.ndim])

  def _quad2res(self, state, f):
    """
      Quadrature results to a field result

      This does Laplacian smoothing
    """
    Md = asm(BilinearForm(lambda u,v,w: v*u), state.sbasis)
    fd = asm(LinearForm(lambda v,w: v*w['values']),
      state.sbasis, values = f)

    return spla.spsolve(Md,fd)

  def _res2quad(self, state, f):
    """
      Results field to the quadrature points
    """
    return state.sbasis.interpolate(f).value
  
  class State:
    """
      Subclass for maintaining state with the python solver
    """
    def __init__(self, tube, mat, qorder, i = None):
      """
        Initialize a full state object
      """
      self.mat = mat
      self.mesh = mesh_tube(tube)
      self.qorder = qorder
      self.ndim = tube.ndim

      self.ri = tube.r - tube.t
      self.ro = tube.r
      self.h = tube.h

      # Define the pressure boundary
      self.define_boundary(tube)
      
      # Base element type
      self.betype = (element.ElementLineP1(), element.ElementQuad1(), 
          element.ElementHex1())[tube.ndim-1]

      # Define the scalar basis for interpolation
      self.define_scalar_basis()

      # Define the interior basis for momentum balance
      self.define_interior_basis()

      # Define the side basis for pressure
      self.define_exterior_basis()
    
      # Now that is out of the way, setup the actual required storage
      self.stress = np.zeros((3,3,self.basis.nelems,self.nqi))
      self.strain = np.zeros((3,3,self.basis.nelems,self.nqi))
      self.mechanical_strain = np.zeros((3,3,self.basis.nelems,self.nqi))
      self.thermal_strain = np.zeros((3,3,self.basis.nelems,self.nqi))
      self.history = np.repeat(self.material.init_store()[:,np.newaxis], 
          self.nq, axis = 1).reshape(
              self.material.nstore,self.basis.nelems, self.nqi)
      self.tangent = np.zeros((3,3,3,3,self.basis.nelems,self.nqi))
      self.temperature = np.zeros((self.basis.nelems,self.nqi))
      self.displacements = self.basis.zeros()
      self.time = 0.0

      self.energy = np.zeros((self.basis.nelems, self.nqi))
      self.dissipation = np.zeros((self.basis.nelems, self.nqi))

      self.force = 0.0
      self.stiffness = 0.0

      if i is not None:
        self.temperature = self.sbasis.interpolate(tube.results['temperature'][i].flatten()).value

    @property
    def material(self):
      """
        Getter for material object
      """
      return get_mat(self.mat)

    def define_boundary(self, tube, tol = 0.5):
      """
        Define the pressure boundary

        Parameters:
          tube      tube object for geometry

        Additional parameters:
          tol       thickness tolerance for finding faces
      """
      atol = tol * tube.t
      if self.ndim == 1:
        self.mesh.define_boundary("pressure", 
            lambda x: np.logical_and(x > tube.r - tube.t - atol, 
              x < tube.r - tube.t + atol))
      else:
        self.mesh.define_boundary("pressure",
            lambda x: np.logical_and(
              np.sqrt(x[0]**2.0+x[1]**2.0) > tube.r - tube.t - atol,
              np.sqrt(x[0]**2.0+x[1]**2.0) < tube.r - tube.t + atol))

    def define_scalar_basis(self):
      """
        Define the scalar basis for mapping around result fields
      """
      self.sbasis = InteriorBasis(mesh = self.mesh,
          elem = self.betype, intorder = self.qorder)

    def define_interior_basis(self):
      """
        Define the interior basis for the balance equation
      """
      etype = element.ElementVectorH1(self.betype)
      self.basis = InteriorBasis(mesh = self.mesh, elem = etype, 
          intorder = self.qorder)

    def define_exterior_basis(self):
      """
        Define the sideset basis for applying pressure
      """
      etype = element.ElementVectorH1(self.betype)
      self.pbasis = FacetBasis(mesh = self.mesh, elem = etype, 
          intorder = self.qorder,
          facets = self.mesh.boundaries['pressure'])

    def copy(self):
      """
        Return a copy

        Soft copy the scikit-fem stuff
        Hard copy the results
      """
      new = copy.copy(self)
      new.stress = np.copy(self.stress)
      new.strain = np.copy(self.strain)
      new.mechanical_strain = np.copy(self.mechanical_strain)
      new.thermal_strain = np.copy(self.thermal_strain)
      new.history = np.copy(self.history)
      new.tangent = np.copy(self.tangent)
      new.temperature = np.copy(self.temperature)
      new.displacements = np.copy(self.displacements)
      new.energy = np.copy(self.energy)
      new.dissipation = np.copy(self.dissipation)
      return new
    
    @property
    def nqi(self):
      """
        Integration points per element
      """
      return len(self.basis.quadrature[1])

    @property
    def nq(self):
      """
        Total number of integration points (size of array)
      """
      return self.nqi * self.basis.mesh.nelements

    @property
    def ne(self):
      """
        Number of elements
      """
      return self.basis.mesh.nelements

def solve_python_1d(state_n, t_n, p_n, state_np1, t_np1, p_np1, top_np1,
    solver_options):
  """
    Solve an increment using the python 1d solver

    Parameters:
      state_n           previous state
      t_n               previous time
      p_n               previous pressure
      state_np1         next state
      t_np1             next time
      p_np1             next pressure
      top_np1           next axial displacement
      solver_options    various solver options
  """
  solver = PythonSolver(state_n, state_np1, solver_options,
      set_strain = top_np1 / state_np1.h)

  solver.solve(t_n, t_np1, p_np1)

def solve_python_2d(state_n, t_n, p_n, state_np1, t_np1, p_np1, top_np1,
    solver_options):
  """
    Solve an increment using the python 2d solver

    Parameters:
      state_n           previous state
      t_n               previous time
      p_n               previous pressure
      state_np1         next state
      t_np1             next time
      p_np1             next pressure
      top_np1           next axial displacement
      solver_options    various solver options
  """
  solver = PythonSolver(state_n, state_np1, solver_options,
      set_strain = top_np1 / state_np1.h)
 
  # Fix the xy line
  node = state_np1.mesh.nodes_satisfying(lambda x:
      np.logical_and(
        np.logical_and(x[0] < state_np1.ro + solver_options['dof_tol'],
          x[0] > state_np1.ro - solver_options['dof_tol']),
        np.logical_and(x[1] < solver_options['dof_tol'],
          x[1] > -solver_options['dof_tol'])))
  dofs_x = state_np1.basis.nodal_dofs[0,node].flatten()
  solver.add_dirichlet_bc(dofs_x, 0.0)
  dofs_y = state_np1.basis.nodal_dofs[1,node].flatten()
  solver.add_dirichlet_bc(dofs_y, 0.0)

  # Fix the y line
  node = state_np1.mesh.nodes_satisfying(lambda x:
      np.logical_and(
        np.logical_and(-x[0] < state_np1.ro + solver_options['dof_tol'],
          -x[0] > state_np1.ro - solver_options['dof_tol']),
        np.logical_and(x[1] < solver_options['dof_tol'],
          x[1] > -solver_options['dof_tol'])))
  dofs_y = state_np1.basis.nodal_dofs[1,node].flatten()
  solver.add_dirichlet_bc(dofs_y, 0.0)

  solver.solve(t_n, t_np1, p_np1)

def solve_python_3d(state_n, t_n, p_n, state_np1, t_np1, p_np1, top_np1,
    solver_options):
  """
    Solve an increment using the python 3d solver

    Parameters:
      state_n           previous state
      t_n               previous time
      p_n               previous pressure
      state_np1         next state
      t_np1             next time
      p_np1             next pressure
      top_np1           next axial displacement
      solver_options    various solver options
  """
  solver = PythonSolver(state_n, state_np1, solver_options)
  
  # Fix the xy line
  node = state_np1.mesh.nodes_satisfying(lambda x:
      np.logical_and(np.logical_and(
        np.logical_and(x[0] < state_np1.ro + solver_options['dof_tol'],
          x[0] > state_np1.ro - solver_options['dof_tol']),
        np.logical_and(x[1] < solver_options['dof_tol'],
          x[1] > -solver_options['dof_tol'])),
        x[2] < solver_options['dof_tol']))
  dofs_x = state_np1.basis.nodal_dofs[0,node].flatten()
  solver.add_dirichlet_bc(dofs_x, 0.0)
  dofs_y = state_np1.basis.nodal_dofs[1,node].flatten()
  solver.add_dirichlet_bc(dofs_y, 0.0)

  # Fix the y line
  node = state_np1.mesh.nodes_satisfying(lambda x:
      np.logical_and(np.logical_and(
        np.logical_and(-x[0] < state_np1.ro + solver_options['dof_tol'],
          -x[0] > state_np1.ro - solver_options['dof_tol']),
        np.logical_and(x[1] < solver_options['dof_tol'],
          x[1] > -solver_options['dof_tol'])),
        x[2] < solver_options['dof_tol']))
  dofs_y = state_np1.basis.nodal_dofs[1,node].flatten()
  solver.add_dirichlet_bc(dofs_y, 0.0)

  # Fix the bottom face
  node = state_np1.mesh.nodes_satisfying(lambda x:
      x[2] < solver_options['dof_tol'])
  dofs = state_np1.basis.nodal_dofs[2,node].flatten()
  solver.add_dirichlet_bc(dofs, 0.0)

  # Fix the top face
  node = state_np1.mesh.nodes_satisfying(lambda x:
      x[2] > state_np1.h - solver_options['dof_tol'])
  dofs = state_np1.basis.nodal_dofs[2,node].flatten()
  solver.add_dirichlet_bc(dofs, top_np1)
  
  solver.solve(t_n, t_np1, p_np1)

class PythonSolver:
  """
    Actually manages the solve with scikit-fem
  """
  def __init__(self, state_n, state_np1, solver_options,
      set_strain = None):
    """
      Setup solver

      Parameters:
        state_n         previous state
        state_np1       current state
        solver_options  various 'how to solve' parameters

      Additional parameters:
        set_strain      set the zz strain to this value, if given
    """
    self.state_n = state_n
    self.state_np1 = state_np1
    self.options = solver_options
    self.set_strain = set_strain

    self.ebcs = []

    self.edofs = []
    self.evalues = []

    # Initialize the guess
    self.setup_guess()

    # The operators!
    if self.ndim == 1:
      # 1D axisymmetric is different than 2D/3D cartesian
      self.internal = LinearForm(lambda v, w: 
          v[1][0][0] * w['radial'] - v[0][0] * w['radial'] / w.x[0] + v[0][0] * w['hoop'] / w.x[0])
      self.jac = BilinearForm(lambda u, v, w:
        v[1][0][0] * w['Crr'] * u[1][0][0] + v[1][0][0] * w['Crt'] * u[0][0] / w.x[0]
        - v[0][0] * (w['Crr'] * u[1][0][0] + w['Crt'] * u[0][0] / w.x[0]) / w.x[0]
        + v[0][0] * (w['Ctr'] * u[1][0][0] + w['Ctt'] * u[0][0] / w.x[0]) / w.x[0])
    else:
      self.internal = LinearForm(lambda v, w: 
          helpers.ddot(helpers.sym_grad(v),w['stress']))
      self.jac = BilinearForm(lambda u,v,w:
          helpers.ddot(helpers.sym_grad(v),
            helpers.ddot(w['C'],helpers.sym_grad(u))))
    self.external = LinearForm(lambda v, w: 
        helpers.dot((-1.0)*w['pressure']*w.n,v))

  @property
  def ndim(self):
    """
      Problem dimension
    """
    return self.state_np1.ndim

  def setup_guess(self):
    """
      Estimate the displacements
    """
    self.state_np1.displacements = self.state_n.displacements

  def solve(self, t_n, t_np1, p):
    """
      Do the actual solve using Newton-Raphson iteration

      Parameters:
        t_n     previous time
        t_np1   next time
        p       pressure
    """
    # Setup the time step
    self.state_n.time = t_n
    self.state_np1.time = t_np1
   
    # Assemble the dirichlet BCs into big vectors
    self.assemble_dirichlet()
    
    # Do the initial stress update
    self.update_state()

    # Calculate the initial residual and jacobian
    R = self.residual(p)
    nR0 = la.norm(R[self.kdofs])

    # Printing, if you want
    if self.options['verbose']:
      print("Iter\tnR\t\tnR/nR0")
      print("%i\t%3.2e" % (0, nR0))
    
    # Newton-Raphson iteration
    for i in range(self.options['miter']):
      # Can immediately skip if the initial residual is 
      # less than the absolute tol
      if nR0 < self.options['atol']:
        if self.options['verbose']:
          print("")
        break
      # Get the actual sparse jacobian
      J = self.jacobian()
      # Update the displacements
      self.state_np1.displacements -= self.linear_solve(J,R)
      # Recalculate the state
      self.update_state()
      # Calculate the residual
      R = self.residual(p)
      # Do the convergence check and print update
      nR = la.norm(R[self.kdofs])
      if self.options['verbose']:
        print("%i\t%3.2e\t%3.2e" % (i+1, nR, nR/nR0))
      if nR < self.options['atol'] or nR / nR0 < self.options['rtol']:
        if self.options['verbose']:
          print("")
        break
    else:
      raise RuntimeError("Nonlinear iteration did not converge!")

    # Calculate the axial force and stiffness
    J = self.jacobian() # Need the current tangent
    if self.state_np1.ndim == 1 or self.state_np1.ndim == 2:
      self.calculate_axial_from_stress(R, J)
    else:
      self.calculate_axial_from_fea(R, J)
  
  def calculate_axial_from_stress(self, R, J):
    """
      Calculate the axial force by integrating sigma_zz over the area

      Parameters:
        R       final residual
        J       final jacobian
    """
    #pylint: disable=no-value-for-parameter, invalid-unary-operand-type
    if self.state_np1.ndim == 1:
      dx = self.state_np1.basis.interpolate(self.state_np1.mesh.p[0]
          )[0][0] * self.state_np1.basis.dx * 2.0 * np.pi
    else:
      dx = self.state_np1.basis.dx

    fake_force = self._internal_force(self.state_np1.tangent[:,:,2,2])
    de = utils.solve(*utils.condense(J, fake_force, D = self.edofs),
        solver = utils.solver_direct_scipy())
    fake_strain = self.calculate_strain(de)
    integrand1 = np.einsum('iijk', self.state_np1.tangent[2,2] * fake_strain)
    integrand2 = self.state_np1.tangent[2,2,2,2]

    self.state_np1.force = np.sum(self.state_np1.stress[2,2] * dx)
    self.state_np1.stiffness = np.sum((-integrand1 + integrand2) * dx) / self.state_np1.h

  def calculate_axial_from_fea(self, R, J):
    """
      Calculate the axial force by summing the discrete forces on the top dofs

      Parameters:
        R       final residual
        J       final jacobian
    """
    node = self.state_np1.mesh.nodes_satisfying(lambda x:
        x[2] > self.state_np1.h - self.options['dof_tol'])
    dofs = self.state_np1.basis.nodal_dofs[2,node].flatten()
    
    # There is a better way to do this
    sdofs = [list(self.edofs).index(d) for d in dofs]
    dotme = np.zeros((len(self.edofs),))
    dotme[sdofs] = 1.0

    K = self.state_np1.basis.complement_dofs(self.edofs)
    
    # Straightforward
    self.state_np1.force = np.sum(R[dofs])
    
    # Expensive but eh can optimize later
    Jl = J.tolil()
    J11 = Jl[K,:][:,K].tocsr() 
    J12 = Jl[K,:][:,self.edofs]

    J21 = Jl[self.edofs,:][:,K]
    J22 = Jl[self.edofs,:][:,self.edofs]
  
    self.state_np1.stiffness = np.dot(dotme, J22.dot(dotme) - 
        J21.dot(spla.spsolve(J11,J12.dot(dotme))))

  def linear_solve(self, J, R):
    """
      Do the linear solve for a particular jacobian and residual

      Parameters:
        J       Jacobian sparse matrix
        R       residual vector
    """
    #pylint: disable=no-value-for-parameter
    return utils.solve(*utils.condense(J, R, D = self.edofs),
        solver = utils.solver_direct_scipy())
  
  def residual(self, p):
    """
      Actually assemble the residual

      Parameters:
        p       Pressure, for the BC
    """
    F_int = self._internal_force(self.state_np1.stress)
    F_ext = asm(self.external, self.state_np1.pbasis, pressure = p)
    
    return F_int - F_ext

  def _internal_force(self, stress):
    """
      Assemble the internal force

      Parameters:
        stress      stresses to use
    """
    if self.ndim == 1:
      return self._internal_1d(stress[0,0], stress[1,1])
    else:
      return self._internal_nd(stress[:self.ndim,:self.ndim])

  def _internal_1d(self, radial, hoop):
    """
      Assemble the internal force for the axisymmetric case

      Parameters:
        radial      radial stress
        hoop        hoop stress
    """
    return asm(self.internal, self.state_np1.basis,
          radial = radial, hoop = hoop) 

  def _internal_nd(self, stress):
    """
      Internal force of the Cartesian cases

      Parameters:
        stress      stresses to use
    """
    return asm(self.internal, self.state_np1.basis,
        stress = stress)

  def jacobian(self):
    """
      Actually assemble the jacobian
    """
    if self.ndim == 1:
      return asm(self.jac, self.state_np1.basis,
          Crr = self.state_np1.tangent[0,0,0,0],
          Crt = self.state_np1.tangent[0,0,1,1],
          Ctt = self.state_np1.tangent[1,1,1,1],
          Ctr = self.state_np1.tangent[1,1,0,0])
    else:
      return asm(self.jac, self.state_np1.basis,
          C = self.state_np1.tangent[:self.ndim,:self.ndim,
            :self.ndim,:self.ndim])

  def add_dirichlet_bc(self, dofs, value):
    """
      Add a dirichlet BC

      Parameters:
        dofs        plain dof list
        value       value to set
    """
    self.ebcs.append((dofs,value))

  def assemble_dirichlet(self):
    """
      Assemble the dirichlet BCs into big vectors
    """
    self.edofs = []
    self.evalues = []

    for d, v in self.ebcs:
      self.edofs.extend(list(d))
      self.evalues.extend([v] * len(d))
    
    self.edofs = np.array(self.edofs, dtype = int)
    self.evalues = np.array(self.evalues)

    inds = np.argsort(self.edofs)
    self.edofs = self.edofs[inds]
    self.evalues = self.evalues[inds]
    self.kdofs = self.state_np1.basis.complement_dofs(self.edofs)

  def update_state(self):
    """
      Make the state consistent with the new displacements
    """
    # Enforce the Dirichlet BCs in the vector
    self.state_np1.displacements[self.edofs] = self.evalues
    
    # Calculate total strain
    self.state_np1.strain = self.calculate_strain(self.state_np1.displacements)

    # Add the optional extra strain
    if self.set_strain is not None:
      self.state_np1.strain[2,2] = self.set_strain

    # Calculate thermal strain and mechanical strain
    self.calculate_mechanical_strain()

    # Calculate stress and tangent
    self.calculate_stress_update()

  def calculate_strain(self, D):
    """
      Calculate the strain based on the current displacements

      Parameters:
        D           displacements to use
    """
    U = self.state_np1.basis.interpolate(D)

    E = np.zeros((3,3) + U[0][0].shape)

    E[:self.ndim,:self.ndim] = helpers.sym_grad(U)
    if self.ndim == 1:
      # Cast needed b/c they don't have a divide operator!
      E[1,1] = np.array(U
          ) / np.array(self.state_np1.basis.interpolate(
            self.state_np1.mesh.p.flatten()))
    
    return E

  def calculate_mechanical_strain(self):
    """
      Calculate the thermal strain increment and update mechanical strain
    """
    # Grab CTE values from NEML
    cte_old = np.zeros(self.state_n.temperature.shape)
    cte_new = np.zeros(self.state_np1.temperature.shape)
    for ind in np.ndindex(self.state_np1.temperature.shape):
      cte_old[ind] = self.state_n.material.alpha(
          self.state_n.temperature[ind])
      cte_new[ind] = self.state_np1.material.alpha(
          self.state_np1.temperature[ind])

    # Thermal strain = thermal_strain_old + alpha * Tdot
    self.state_np1.thermal_strain = self.state_n.thermal_strain + np.eye(
        3)[:,:,np.newaxis,np.newaxis] * (
        (cte_new+cte_old)/2.0*(self.state_np1.temperature - 
      self.state_n.temperature))

    # Mechanical strain = total - thermal
    self.state_np1.mechanical_strain = (self.state_np1.strain -
        self.state_np1.thermal_strain)

  def calculate_stress_update(self):
    """
      The expensive function: update the stress and tangent
    """
    # Ya, you can't a view of these arrays...
    # We could reorder everything and then it would just work natively
    stress_np1 = np.zeros((self.state_np1.nq,3,3))
    hist_np1 = np.zeros((self.state_np1.nq,self.state_np1.material.nstore))
    A_np1 = np.zeros((self.state_np1.nq,3,3,3,3))

    #pylint: disable=c-extension-no-member
    block.block_evaluate(self.state_np1.material,
        self.state_np1.mechanical_strain.transpose(2,3,0,1).reshape(-1,3,3), 
        self.state_n.mechanical_strain.transpose(2,3,0,1).reshape(-1, 3,3),
        self.state_np1.temperature.flatten(), 
        self.state_n.temperature.flatten(),
        self.state_np1.time, self.state_n.time,
        stress_np1, self.state_n.stress.transpose(2,3,0,1).reshape(-1,3,3),
        hist_np1, 
        self.state_n.history.transpose(1,2,0).reshape(
          -1,self.state_n.material.nstore),
        A_np1,
        self.state_np1.energy.flatten(), 
        self.state_n.energy.flatten(),
        self.state_np1.dissipation.flatten(),
        self.state_n.dissipation.flatten())
    
    self.state_np1.stress = stress_np1.reshape((self.state_np1.ne,
      self.state_np1.nqi, 3, 3)).transpose(2,3,0,1)
    self.state_np1.history = hist_np1.reshape((self.state_np1.ne,
      self.state_np1.nqi, self.state_np1.material.nstore)
        ).transpose(2,0,1)
    self.state_np1.tangent = A_np1.reshape((self.state_np1.ne,
      self.state_np1.nqi, 3,3,3,3)).transpose(2,3,4,5,0,1)
   
def get_mat(thing):
  """
    Small helper to wrap NEML for pickling issues
  """
  try:
    return thing.get_neml_model()
  except AttributeError:
    return thing



"""
  Module that facilitates loading in all the material data
"""
import os.path

from srlife import materials

LIBRARY_DIR = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'data'))

def get_file(directory, name):
  """ Return a file path or raise an error

    Args:
      directory (str): directory the file should be in
      name (str): actual file name

    Raises:
      RunetimeError: if the file does not exist
  """
  attempt = os.path.join(directory, name+".xml")
  if not os.path.exists(attempt):
    raise RuntimeError("Material with name %s does not exists in database!" 
        % name)
  return attempt

def load_fluid(name, model):
  """ Load fluid material properties:

    Args:
      name (str): name of the fluid (title of xml file)
      model (str):  particular convection model to use

    Returns:
      material.FluidMaterial: fluid material model
  """
  fdir = os.path.join(LIBRARY_DIR, "fluid")
  filename = get_file(fdir, name)
  return materials.FluidMaterial.load(filename, model)

def load_material(name, thermal_model, deformation_model, damage_model):
  """ Load solid material properties

    Args:
      name (str): name of the material (title of xml file)
      thermal_model (str): which thermal model variant to use
      deformation_model (str): which deformation model variant to use
      damage_model (str): which damage model variant to use

    Returns:
      materials.ThermalMaterial: thermal material model
      materials.DeformationMaterial: deformation material model
      materials.StructuralMaterial: damage material model
  """
  return load_thermal(name, thermal_model), load_deformation(
      name, deformation_model), load_damage(name, damage_model)

def load_thermal(name, model):
  """ Load thermal material data

    Args:
      name: name of the material
      model: model variant

    Returns:
      materials.ThermalMaterial: thermal material model
  """
  fdir = os.path.join(LIBRARY_DIR, "thermal")
  filename = get_file(fdir, name)
  return materials.ThermalMaterial.load(filename, model)

def load_deformation(name, model):
  """ Load a deformation model from file

    Args:
      name: name of the material
      model: model variant

    Returns:
      materials.DeformationMaterial: deformation material model
  """
  fdir = os.path.join(LIBRARY_DIR, "deformation")
  filename = get_file(fdir, name)
  return materials.DeformationMaterial(filename, model)

def load_damage(name, model):
  """ Load a damage model from file

    Args:
      name: name of the material
      model: model variant

    Returns:
      materials.StructuralMaterial: damage material model
  """
  fdir = os.path.join(LIBRARY_DIR, "damage")
  filename = get_file(fdir, name)
  return materials.StructuralMaterial.load(filename, model)



#pylint: disable=no-member

"""
  Meta-structural solver defining a network of nonlinear springs
"""

from abc import ABC, abstractmethod

import numpy as np
import networkx as nx
import multiprocess

from srlife.solvers import newton

class Spring(ABC):
  """
    Spring base class defining interface.

    Class will accommodate full nonlinear tubes (through FEA) as well 
    as simple linear springs
  """
  @abstractmethod
  def force_and_stiffness(self, i, d):
    """
      Return the current forces and stiffness for some displacement

      Parameters:
        i           timestep
        d           displacement
    """
    return

  @abstractmethod
  def update_state(self, i):
    """
      Update to the next state after a successful solve

      Parameters:
        i           universal timestep number
    """
    return

class LinearSpring(Spring):
  """
    A linear elastic spring
  """
  def __init__(self, k):
    """
      Parameters:
        k       spring stiffness
    """
    self.k = k

  def force_and_stiffness(self, i, d):
    """
      Simple linear update

      Parameters:
        i       universal timestep index
        d       current displacement
    """
    return self.k*d, self.k

  def update_state(self, i):
    """
      We don't have state!

      Parameters:
        i       universal timestep number
    """
    return

class TubeSpring(Spring):
  """
    A spring object encapsulating the entire tube thermo-structural solver
  """
  def __init__(self, tube, solver, material):
    """
      Parameters:
        tube        tube object, with thermal history defined
        solver      structural solver to use
        material    tube material model (NEML)
    """
    self.tube = tube
    self.solver = solver
    self.material = material

    # Setup the tube to receive results
    self.solver.setup_tube(self.tube)

    # Get the first state
    self.state_n = self.solver.init_state(self.tube, self.material, 
        i = 0)

    # Dump this first state to the Tube
    self.solver.dump_state(self.tube, 0, self.state_n)

  def force_and_stiffness(self, i, d):
    """
      Parameters:
        d       current displacement
    """
    self.state_np1 = self.solver.solve(self.tube, i, self.state_n, d)

    return self.state_np1.force, self.state_np1.stiffness

  def update_state(self, i):
    """
      Parameters:
        i       timestep
    """
    self.solver.dump_state(self.tube, i, self.state_np1)
    self.state_n = self.state_np1

class SpringNetwork(nx.MultiGraph):
  """
    This mimics the NEML class with some additional features.

    Graph structure:
      Nodes are displacements
      Edges can be:
        Bars
        "disconnect"
        "rigid"

      Where "free" and "fixed" get factored out in the final network

    Boundary conditions:
      Live at nodes
      Can be imposed forces or displacements (though only displacements are
        going to bused in the end)
      Are simply tuples of (type, function) where 
        type:
          force
          displacement
        function:
          function of time giving the values

    The object takes the time values from individual tubes by default, but
    this can be overrode with new step values.
  """
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.times = None
    self.displacements = None
    self.atol = kwargs.pop("atol", 1.0e-8)
    self.rtol = kwargs.pop("rtol", 1.0e-6)
    self.miter = kwargs.pop("miter", 25)
    self.verbose = kwargs.pop("verbose", False)

  def __copy(self, other):
    """
      Copy ancillary information from one network to another

      Parameters:
        other       other network
    """
    other.times = self.times
    other.rtol = self.rtol
    other.atol = self.atol
    other.miter = self.miter
    other.verbose = self.verbose

  def set_times(self, times):
    """
      Set the discrete times to a list

      Parameters:
        times       new times
    """
    self.times = times

  def validate_setup(self):
    """
      Validate the data structure 
    """
    if self.times is None:
      self.__set_default_times()

    # Validate the type of each edge and the internal times (if provided)
    for _,_,edge in self.edges(data=True):
      if isinstance(edge['object'], TubeSpring):
        if not np.allclose(edge['object'].tube.times, self.times):
          raise ValueError("A tube object has discrete times that are"
              " not consistent with the spring network!")
      
      if not isinstance(edge['object'], (Spring,str)):
        raise ValueError("An edge object is not either a spring or a string!")

      if isinstance(edge['object'], str) and edge['object'] not in (
          "disconnect", "rigid"):
        raise ValueError("Special edge type %s not recognized!" % edge)

  def __set_default_times(self):
    """
      Grab times from the first Tube object or raise error
    """
    for _,_,edge in self.edges(data=True):
      if isinstance(edge['object'], TubeSpring):
        self.times = edge['object'].tube.times
        break
    else:
      raise ValueError("No times provided and no Tube objects to copy from!")

  def force_bc(self, node, function):
    """
      Add a force boundary condition

      Parameters:
        node        location
        function    force as a function of time
    """
    self.nodes[node]['bc'] = ('force', function)

  def displacement_bc(self, node, function):
    """
      Add a displacement boundary condition

      Parameters:
        node        location
        function    displacement as a function of time
    """
    self.nodes[node]['bc'] = ('displacement', function)

  def reduce_graph(self):
    """
      Reduce the graph structure to a list of graphs which can be
      solved by:

      1) Removing rigid links
      2) Splitting the graph at disconnect
    """
    self.remove_rigid()
    return self.split_disconnect()

  def remove_rigid(self):
    """
      Remove rigid links by merging the nodes in place
    """
    while True:
      for i,j, edge in self.edges(data=True):
        # Can't handle duplicate springs
        if edge['object'] == "rigid":
          if self.number_of_edges(i,j) != 1:
            raise RuntimeError("Cannot have a rigid link across a spring!")
          bci = 'bc' in self.nodes[i]
          bcj = 'bc' in self.nodes[j]
          if bci and bcj:
            raise RuntimeError("Cannot merge two nodes with BCs!")
          self.remove_edge(i,j)
          if bci or ((not bci) and (not bcj)):
            self.replace(nx.contracted_nodes(self, i, j))
          else:
            self.replace(nx.contracted_nodes(self, j, i))
          break
      else:
        break

  def replace(self, other):
    """
      Replace the graph data structure with a new graph
    """
    self.clear()
    self.update(other)

  def split_disconnect(self):
    """
      Split the graph into disjoint parts by taking apart at the disconnects
    """
    while True:
      for i,j,key,edge in self.edges(data=True, keys = True):
        if edge['object'] == "disconnect":
          self.remove_edge(i,j, key = key)
          break
      else:
        break
    
    components = list(nx.connected_components(self))

    nobjs = [SpringNetwork(self.subgraph(nset)) for nset in components]
    for obj in nobjs:
      self.__copy(obj)

    fobjs = []
    for obj in nobjs:
      if obj.size() != 0:
        fobjs.append(obj)

    return fobjs

  def validate_solve(self):
    """
      Validate that the graph is ready to actually solve

      Conditions:
        1) All dummy edges removed
        2) At least one fixed condition
    """
    # Check that we have times
    if self.times is None:
      raise RuntimeError("Discrete times must be established at solve!")

    # Check that everything is a spring
    for _,_,edge in self.edges(data=True):
      if not isinstance(edge['object'], Spring):
        raise RuntimeError("All edges must be springs at solve!")
    
    # Check that there is at least one fixed boundary
    ndisp = 0
    for node in self.nodes():
      if 'bc' in self.nodes[node]:
        if self.nodes[node]['bc'][0] == 'displacement':
          ndisp += 1

    if ndisp == 0:
      raise RuntimeError("Spring network requires at least one fixed BC!")

    # To to see if the graph is fully-connected
    if not nx.is_connected(self):
      raise RuntimeError("Spring network must be fully connected at solve!")

  def solve(self, i, nthreads = 1):
    """
      Solve discrete step i

      Parameter:
        i           discrete time step

      Additional parameters:
        nthreads    number of threads to use
    """
    # Keep forces and displacements for debug
    if self.displacements is None:
      self.displacements = np.zeros((len(self.nodes),))

    # Make sure we can solve
    self.validate_solve()

    # Set the time step
    self.i = i
    
    # Figure out free/fixed displacements
    (self.dmap, self.free, self.forces, self.fixed,
        self.fixed_displacements) = self.dof_maps(i)

    # Actually solve
    d = newton(lambda x: self.RJ(x, nthreads = nthreads),
        self.displacements[self.dmap[self.free]], verbose = self.verbose,
        rel_tol = self.rtol, abs_tol = self.atol, miters = self.miter)

    # Store the displacements, for fun
    self.displacements[self.dmap[self.free]] = d
    self.displacements[self.dmap[self.fixed]] = self.fixed_displacements

    # Advance the state
    for _,_,edge in self.edges(data=True):
      edge['object'].update_state(self.i)

  def RJ(self, d, nthreads = 1):
    """
      Actually calculate the residual and Jacobian equation for the step
      
      Parameters:
        d       free displacements
    """
    dall = np.zeros((len(self.nodes),))
    dall[self.dmap[self.fixed]] = self.fixed_displacements
    dall[self.dmap[self.free]] = d
    
    if nthreads > 1:
      with multiprocess.Pool(nthreads) as p:
        res = list(p.map(lambda e: self.fj(dall, *e),  self.edges(data=True)))
    else:
      res = list(map(lambda e: self.fj(dall, *e),  self.edges(data=True)))
    Fint = sum(r[0] for r in res)
    J = sum(r[1] for r in res)

    for k,(_,_,edge) in enumerate(self.edges(data=True)):
      edge['object'].state_np1 = res[k][2]

    return (Fint[self.dmap[self.free]] - self.forces, 
        J[self.dmap[self.free],:][:,self.dmap[self.free]])

  def fj(self,dall,i,j,edge):
    """
      Calculate the force and jacobian contributions for a single edge 
    """
    Fint = np.zeros((len(self.nodes),))
    J = np.zeros((len(self.nodes),len(self.nodes)))
    f, k = edge['object'].force_and_stiffness(self.i, dall[self.dmap[j]] 
        - dall[self.dmap[i]])
    Fint[self.dmap[i]] += -f
    Fint[self.dmap[j]] += f
    
    J[self.dmap[i],self.dmap[i]] += k
    J[self.dmap[i],self.dmap[j]] += -k
    J[self.dmap[j],self.dmap[i]] += -k
    J[self.dmap[j],self.dmap[j]] += k

    return Fint, J, edge['object'].state_np1

  def dof_maps(self, i):
    """
      Get the degree of freedom maps for the given time step

      Parameters:
        i       discrete time step
    """
    free = []
    fixed = []
    forces = []
    displacements = []
    time = self.times[i]
    for node in self.nodes:
      if 'bc' in self.nodes[node]:
        if self.nodes[node]['bc'][0] == 'displacement':
          fixed.append(node)
          displacements.append(self.nodes[node]['bc'][1](time))
        elif self.nodes[node]['bc'][0] == 'force':
          free.append(node)
          forces.append(self.nodes[node]['bc'][1](time))
        else:
          raise ValueError("Unknown BC type %s!" % self.nodes[node]['bc'][0])
      else:
        free.append(node)
        forces.append(0)

    nodes = np.array(self.nodes, dtype=int)
    total = (np.array(range(0, max(nodes)+1), dtype = int)*0)-1
    rnums = np.array(range(len(nodes)), dtype = int)
    total[nodes] = rnums

    return total,np.array(free, dtype = int), np.array(forces), np.array(fixed,
        dtype = int), np.array(displacements)

  def solve_all(self, nthreads = 1, decorator = lambda x, nitems: x):
    """
      Solve all time steps

      Additional parameters:
        nthreads:       number of threads to use
        decorator:      progress bar decorator
    """
    self.validate_solve()
    list(decorator(
      map(lambda i: self.solve(i, nthreads), range(1, len(self.times))), 
      len(self.times)-1))
    return self



import unittest
import tempfile
import os, os.path

import numpy as np

from srlife import receiver, thermal, materials

def make_tube(D, period):
  R = 1.0
  t = 0.1
  h = 10.0
  nr = 10
  nt = 20
  nz = 10
  T0 = 50

  tmax = period

  ntime = 10

  T0 = 300
  
  tube = receiver.Tube(R, t, h, nr, nt, nz, T0 = T0)
  if D == 1:
    tube.make_1D(h/2,1)
  elif D == 2:
    tube.make_2D(h/2)

  times = np.linspace(0,tmax, ntime)

  tube.set_times(times)

  Tf_0 = 300
  Tf_m = 600

  def fluid_T(t):
    if t < tmax / 2.0:
      return Tf_0 + (Tf_m - Tf_0) / (tmax/2.0) * t
    else:
      return Tf_m - (Tf_m - Tf_0) / (tmax/2.0) * (t - tmax/2.0)

  ftemps = np.array([fluid_T(t) * np.ones((nz,)) for t in times])

  inner_convection = receiver.ConvectiveBC(R - t, h, nz, 
      times, ftemps)
  tube.set_bc(inner_convection, "inner")

  hflux = receiver.HeatFluxBC(R, h, nt, nz, times, 
      np.zeros((ntime,nt,nz))+2.0)
  tube.set_bc(hflux, "outer")

  solver = thermal.FiniteDifferenceImplicitThermalSolver()

  tmat = materials.ConstantThermalMaterial("dummy", 20.0e-3, 4.8)
  fmat = materials.ConstantFluidMaterial({"dummy": 8.1e-3})

  solver.solve(tube, tmat, fmat)

  return tube

def make_panel(D, ntube, period, stiff):
  panel = receiver.Panel(stiff)

  # To save some time
  t = make_tube(D, period)
  
  for i in range(ntube):
    panel.add_tube(t)

  return panel

def make_receiver(D, npanel, ntube, period = 8 * 3600.0, days = 1, 
    panel_stiffness = 0.0, manifold_stiffness = 0.0):
  r = receiver.Receiver(period, days, panel_stiffness)
  for i in range(npanel):
    r.add_panel(make_panel(D, ntube, period, manifold_stiffness))

  return r


class TestVTKWriter(unittest.TestCase):
  def setUp(self):
    self.period = 8 * 3600.0
    self.npanel = 2
    self.ntube = 3
    self.days = 1
    self.panel_stiffness = 100.0
    self.manifold_stiffness = 100.0

  def test_1D(self):
    r = make_receiver(1, self.npanel, self.ntube, period = self.period,
        days = self.days, 
        panel_stiffness = self.panel_stiffness,
        manifold_stiffness = self.manifold_stiffness)
    tdir = tempfile.mkdtemp()

    r.write_vtk(os.path.join(tdir, "test"))

    nfile = len(os.listdir(tdir))

    self.assertEqual(nfile, self.npanel * self.ntube)

  def test_2D(self):
    r = make_receiver(2, self.npanel, self.ntube, period = self.period,
        days = self.days,
        panel_stiffness = self.panel_stiffness,
        manifold_stiffness = self.manifold_stiffness)
    tdir = tempfile.mkdtemp()

    r.write_vtk(os.path.join(tdir, "test"))

    nfile = len(os.listdir(tdir))

    self.assertEqual(nfile, self.npanel * self.ntube)

  def test_3D(self):
    r = make_receiver(2, self.npanel, self.ntube, period = self.period,
        days = self.days, 
        panel_stiffness = self.panel_stiffness,
        manifold_stiffness = self.manifold_stiffness)
    tdir = tempfile.mkdtemp()

    r.write_vtk(os.path.join(tdir, "test"))

    nfile = len(os.listdir(tdir))

    self.assertEqual(nfile, self.npanel * self.ntube)




import unittest

import numpy as np

from srlife import materials, thermal, receiver

import sys
sys.path.append("test/solvers/thermal")
from thermalsol import ManufacturedSolution

class TestThermalManufactured(unittest.TestCase):
  def setUp(self):
    self.problems = [
        ManufacturedSolution("1D: no spatial", 1,
          lambda t, r: t,
          lambda t, k, alpha, r: k/alpha * (r*0.0+1)),
        ManufacturedSolution("1D: spatial and time", 1,
          lambda t, r: np.sin(t)*np.log(r),
          lambda t, k, alpha, r: k/alpha * np.log(r) * np.cos(t)),
        ManufacturedSolution("1D: just spatial", 1,
          lambda t, r: np.sin(r),
          lambda t, k, alpha, r: k * np.sin(r) - k/r*np.cos(r)),
        ManufacturedSolution("2D: just time", 2,
          lambda t, r, th: t,
          lambda t, k, alpha, r, th: k/alpha * (r*0.0+1)),
        ManufacturedSolution("2D: just r", 2,
          lambda t, r, th: np.sin(r),
          lambda t, k, alpha, r, th: k * np.sin(r) - k/r * np.cos(r)),
        ManufacturedSolution("2D: just theta", 2,
          lambda t, r, th: np.cos(th),
          lambda t, k, alpha, r, th: k * np.cos(th) / r**2.0),
        ManufacturedSolution("2D: theta and r", 2,
          lambda t, r, th: np.cos(th) / r,
          lambda t, k, alpha, r, th: -k*np.cos(th) / r**3.0),
        ManufacturedSolution("2D: all three", 2,
          lambda t, r, th: np.log(r) * np.sin(th) / (t+1),
          lambda t, k, alpha, r, th: k*np.log(r)*np.sin(th)/((t+1)*r**2.0) - k/alpha * np.log(r) * np.sin(th) / (t+1)**2.0),
        ManufacturedSolution("3D: just t", 3,
          lambda t, r, th, z: t,
          lambda t, k, alpha, r, th, z: k/alpha * (r*0.0+1)),
        ManufacturedSolution("3D: just r", 3,
          lambda t, r, th, z: np.sin(r),
          lambda t, k, alpha, r, th, z: k * np.sin(r) - k/r * np.cos(r)),
        ManufacturedSolution("3D: just theta", 3,
          lambda t, r, th, z: np.cos(th),
          lambda t, k, alpha, r, th, z: k * np.cos(th) / r**2.0),
        ManufacturedSolution("3D: just z", 3,
          lambda t, r, th, z: np.sin(z),
          lambda t, k, alpha, r, th, z: k * np.sin(z)),
        ManufacturedSolution("3D: all spatial", 3,
          lambda t, r, th, z: z**2.0*np.cos(th)/r,
          lambda t, k, alpha, r, th, z: -k*np.cos(th)/r*((z/r)**2.0+2)),
        ManufacturedSolution("3D: everything", 3,
          lambda t, r, th, z: np.log(r)*np.sin(th)*np.cos(z)/(t+1.0),
          lambda t, k, alpha, r, th, z: k*np.log(r) * np.sin(th) * np.cos(z) / (t+1) * (1.0 + 1/r**2.0 - 1.0/(alpha*(t+1)))),
        ]

    self.solver = thermal.FiniteDifferenceImplicitThermalSolver()
    self.material = materials.ConstantThermalMaterial("Test", 10.0, 5.0)
    self.fluid = materials.ConstantFluidMaterial({"Test": 7.5})

    self.tol = 1e-2
    self.atol = 1e-3

  def _check_case(self, case):
    res = case.solve(self.solver, self.material, self.fluid)
    if not case.assess_comparison(res, self.tol, self.atol):
      print(case.name)
      self.assertTrue(False)

  def test_all_problems(self):
    for i,case in enumerate(self.problems):
      self._check_case(case)

class TestThermalBCs(unittest.TestCase):
  def setUp(self):
    self.k = 15.0
    self.a = 5.0
    self.hcoef = 1.0

    self.solver = thermal.FiniteDifferenceImplicitThermalSolver()
    self.material = materials.ConstantThermalMaterial("Test", self.k, self.a)
    self.fluid = materials.ConstantFluidMaterial({"Test": self.hcoef})

    self.t = 0.2
    self.r = 2.0
    self.h = 1.0
    self.nr = 10

    self.nt = 200
    self.nz = 2

    self.tmax = 1000.0
    self.ntime = 10

    self.tube = receiver.Tube(self.r, self.t, self.h, self.nr, self.nt, self.nz, 0)

    self.times = np.linspace(0, self.tmax, self.ntime+1)
    self.tube.set_times(self.times)

    self.tube.make_1D(self.tube.h/2, 0)

    self.ttimes, self.thetas, self.zs = np.meshgrid(
        self.times, np.linspace(0, 2*np.pi, self.nt + 1)[:-1],
        np.linspace(0, self.h, self.nz), indexing = 'ij')

    self.rs = np.linspace(self.r - self.t, self.r, self.nr)

  def test_dirichlet(self):
    T0 = lambda x: np.sin(2*np.pi*(x-self.r-self.t)/self.t)

    Tleft = receiver.FixedTempBC(self.r-self.t, self.h, self.nt, self.nz,
        self.times, np.zeros(self.ttimes.shape))
    self.tube.set_bc(Tleft, "inner")

    Tright = receiver.FixedTempBC(self.r, self.h, self.nt, self.nz,
        self.times, np.zeros(self.ttimes.shape))
    self.tube.set_bc(Tright, "outer")

    self.solver.solve(self.tube, self.material, self.fluid,
        T0 = T0)

    # Correct solution: temperatures head towards zero
    T = self.tube.results['temperature'][-1]
    self.assertTrue(np.allclose(T,0))

  def test_neumann_left(self):
    T0 = lambda x: np.sin(2*np.pi*(x-self.r-self.t)/self.t)
    q = 10.0

    Tleft = receiver.HeatFluxBC(self.r-self.t, self.h, self.nt, self.nz,
        self.times, np.ones(self.ttimes.shape)*q)
    self.tube.set_bc(Tleft, "inner")

    Tright = receiver.FixedTempBC(self.r, self.h, self.nt, self.nz,
        self.times, np.zeros(self.ttimes.shape))
    self.tube.set_bc(Tright, "outer")

    self.solver.solve(self.tube, self.material, self.fluid,
        T0 = T0)

    # Correct solution:
    Tright = -q * (self.r - self.t) / self.k * np.log(self.rs) + q * (
        self.r - self.t) / self.k * np.log(self.r)

    T = self.tube.results['temperature'][-1]

    self.assertTrue(np.allclose(T, Tright, rtol = 1e-2))

  def test_neumann_right(self):
    T0 = lambda x: np.sin(2*np.pi*(x-self.r-self.t)/self.t)
    q = 10.0

    Tright = receiver.HeatFluxBC(self.r, self.h, self.nt, self.nz,
        self.times, np.ones(self.ttimes.shape)*q)
    self.tube.set_bc(Tright, "outer")

    Tleft = receiver.FixedTempBC(self.r - self.t, self.h, self.nt, self.nz,
        self.times, np.zeros(self.ttimes.shape))
    self.tube.set_bc(Tleft, "inner")

    self.solver.solve(self.tube, self.material, self.fluid,
        T0 = T0)

    # Correct solution:
    Tright = q * self.r / self.k * np.log(self.rs) - q * self.r / self. k * np.log(
        self.r - self.t)

    T = self.tube.results['temperature'][-1]

    self.assertTrue(np.allclose(T, Tright, rtol = 1e-2))

  def test_convection_left(self):
    T0 = lambda x: np.sin(2*np.pi*(x-self.r-self.t)/self.t)

    T_inner = 25

    Tin = receiver.ConvectiveBC(self.r-self.t, self.h, self.nz,
        self.times, np.ones((self.ntime+1,self.nz))*T_inner)
    self.tube.set_bc(Tin, "inner")

    Tright = receiver.FixedTempBC(self.r, self.h, self.nt, self.nz,
        self.times, np.zeros(self.ttimes.shape))
    self.tube.set_bc(Tright, "outer")

    self.solver.solve(self.tube, self.material, self.fluid,
        T0 = T0)

    T = self.tube.results['temperature'][-1]

    ri = self.r - self.t

    C1 = T_inner * self.hcoef / (self.hcoef*np.log(ri) - self.hcoef*np.log(self.r) - self.k/ri)
    C2 = -C1 * np.log(self.r)

    Texact = C1 * np.log(self.rs) + C2

    self.assertTrue(np.allclose(T,Texact, rtol = 1.0e-2))

  def test_convection_right(self):
    T0 = lambda x: np.sin(2*np.pi*(x-self.r-self.t)/self.t)

    T_outer = 25

    Tout = receiver.ConvectiveBC(self.r, self.h, self.nz,
        self.times, np.ones((self.ntime+1,self.nz))*T_outer)
    self.tube.set_bc(Tout, "outer")

    Tleft = receiver.FixedTempBC(self.r-self.t, self.h, self.nt, self.nz,
        self.times, np.zeros(self.ttimes.shape))
    self.tube.set_bc(Tleft, "inner")

    self.solver.solve(self.tube, self.material, self.fluid,
        T0 = T0)

    T = self.tube.results['temperature'][-1]

    ri = self.r - self.t

    C1 = self.hcoef * T_outer / (self.k/self.r + self.hcoef * np.log(self.r) - self.hcoef * np.log(ri))
    C2 = -C1 * np.log(ri)

    Texact = C1 * np.log(self.rs) + C2

    self.assertTrue(np.allclose(T,Texact, rtol = 1.0e-2))

  def test_substep(self):
    T0 = lambda x: np.sin(2*np.pi*(x-self.r-self.t)/self.t)

    T_outer = 250

    Tout = receiver.ConvectiveBC(self.r, self.h, self.nz,
        self.times, np.ones((self.ntime+1,self.nz))*T_outer)
    self.tube.set_bc(Tout, "outer")

    Tleft = receiver.FixedTempBC(self.r-self.t, self.h, self.nt, self.nz,
        self.times, np.zeros(self.ttimes.shape))
    self.tube.set_bc(Tleft, "inner")

    self.solver.solve(self.tube, self.material, self.fluid,
        T0 = T0, substep = 10)

    T = self.tube.results['temperature'][-1]

    ri = self.r - self.t

    C1 = self.hcoef * T_outer / (self.k/self.r + self.hcoef * np.log(self.r) - self.hcoef * np.log(ri))
    C2 = -C1 * np.log(ri)

    Texact = C1 * np.log(self.rs) + C2

    self.assertTrue(np.allclose(T,Texact, rtol = 1.0e-2))

class TestFunction(unittest.TestCase):
  def setUp(self):
    self.k = 15.0
    self.a = 5.0
    self.hcoef = 1.0

    self.solver = thermal.FiniteDifferenceImplicitThermalSolver()
    self.material = materials.ConstantThermalMaterial("Test", self.k, self.a)
    self.fluid = materials.ConstantFluidMaterial({"Test": self.hcoef})

    self.t = 0.2
    self.r = 2.0
    self.h = 1.0

    self.nr = 5

    self.nt = 10
    self.nz = 5

    self.tmax = 10.0
    self.ntime = 10

    self.tube = receiver.Tube(self.r, self.t, self.h, self.nr, self.nt, self.nz, 0)

    self.times = np.linspace(0, self.tmax, self.ntime+1)
    self.tube.set_times(self.times)

    self.T_inner = 100

    Tin = receiver.ConvectiveBC(self.r-self.t, self.h, self.nz,
        self.times, np.ones((self.ntime+1,self.nz))*self.T_inner)
    self.tube.set_bc(Tin, "inner")

    Tright = receiver.HeatFluxBC(self.r, self.h, self.nt, self.nz,
        self.times, np.zeros((self.ntime+1,self.nt,self.nz)))
    self.tube.set_bc(Tright, "outer")

  def test_1D(self):
    self.tube.make_1D(self.tube.h/2, 0)
    self.solver.solve(self.tube, self.material, self.fluid)

  def test_2D(self):
    self.tube.make_2D(self.tube.h/2)
    self.solver.solve(self.tube, self.material, self.fluid)

  def test_3D(self):
    self.solver.solve(self.tube, self.material, self.fluid)



"""
  Solution managers actually walk a model through all the steps to solve
"""
import multiprocess
import tqdm

from srlife import solverparams

class SolutionManager:
  """ Solution manager

    High level solution manager walking through the thermal, structural,
    and damage calculations.

    Args:
      receiver (receiver.Receiver): receiver object to solve
      thermal_solver (thermal.ThermalSolver): how to solve the heat transport
      thermal_material (materials.ThermalMaterial): solid thermal properties
      fluid_material (materials.FluidMaterial): fluid thermal properties
      structural_solver (structural.TubeSolver):  how to solve the mechanics problem
      deformation_material (materials.DeformationMaterial):  how things deform with time
      damage_material (materials.StructuralMaterial):  how to calculate creep damage
      system_solver (system.SystemSolver):  how to tie tubes into the structural system
      damage_model (damage.DamageCalculator): how to calculate damage from the results
      pset (Optional[solverparams.ParameterSet]): optional set of solver parameters
  """
  def __init__(self, receiver, thermal_solver, thermal_material,
      fluid_material, structural_solver, deformation_material,
      damage_material, system_solver, damage_model,
      pset = solverparams.ParameterSet()):
    self.receiver = receiver
    self.thermal_solver = thermal_solver
    self.thermal_material = thermal_material
    self.fluid_material = fluid_material
    self.structural_solver = structural_solver
    self.deformation_material = deformation_material
    self.damage_material = damage_material
    self.system_solver = system_solver
    self.damage_model = damage_model

    self.pset = pset
    self.nthreads = pset.get_default("nthreads", 1)
    self.progress = pset.get_default("progress_bars", False)

  @property
  def tubes(self):
    """ Direct iterator over tubes

    Returns:
      iterator over tubes
    """
    return self.receiver.tubes

  @property
  def ntubes(self):
    """ Pass through to get the number of tubes that need to be analyzed

    Returns:
      int:  total number of tubes in entire receiver
    """
    return len(list(self.tubes))
  
  def progress_decorator(self, base, ntotal):
    """ Either wrap with a progress bar decorator or return a dummy
      
      Args:
        base (function): base function to wrap
        ntotal (int): total number in iterator, needed for wrapping iterators

      Returns:
        either function wrapped with decorator or base function
    """
    if self.progress:
      return tqdm.tqdm(base, total = ntotal)
    else:
      return base

  def solve_life(self):
    """ User interface: solve everything and return receiver life

      The trigger for everything: solve the complete problem and report the
      best-estimate life.

      Returns:
        float:  Number of allowed daily cycles
    """
    self.solve_heat_transfer()
    self.solve_structural()

    return self.calculate_damage()

  def calculate_damage(self):
    """ Calculate damage from the results

    Returns:
      float:    Number of allowed daily cycles
    """
    if self.progress:
      print("Calculating damage:")
    return self.damage_model.determine_life(self.receiver, self.damage_material, 
        nthreads = self.nthreads, decorator = self.progress_decorator)

  def solve_heat_transfer(self):
    """ Solve the heat transfer problem for each tube

    Adds the thermal results to each receiver.Tube object
    """
    #pylint: disable=no-member
    if self.progress:
      print("Running thermal analysis:")
    with multiprocess.Pool(self.nthreads) as p:
      temps = list(
          self.progress_decorator(
          p.imap(lambda x: self.thermal_solver.solve(x, self.thermal_material, self.fluid_material),
        self.tubes), self.ntubes)
          )
    for tube, temps in zip(self.tubes, temps):
      tube.add_results("temperature", temps)

  def solve_structural(self):
    """ Solve the structural problem for the complete system

    Adds the structural results to each tube.
    """
    if self.progress:
      print("Running structural analysis:")
    self.system_solver.solve(self.receiver, self.deformation_material,
        self.structural_solver, nthreads = self.nthreads, 
        decorator = self.progress_decorator, verbose = self.progress)



"""
  Helper functions for converting to and from tensor notation
"""

from __future__ import division

import itertools

import numpy as np

mandel = ((0,0),(1,1),(2,2),(1,2),(0,2),(0,1))
mandel_mults = (1,1,1,np.sqrt(2),np.sqrt(2),np.sqrt(2))

def make_M2T():
  """
    Make the tensor that goes from rank 4 Mandel to rank 4 tensors
  """
  R = np.zeros((3,3,3,3,6,6))
  for a in range(6):
    for b in range(6):
      ind_a = itertools.permutations(mandel[a], r=2)
      ind_b = itertools.permutations(mandel[b], r=2)
      ma = mandel_mults[a]
      mb = mandel_mults[b]
      indexes = tuple(ai+bi for ai, bi in itertools.product(ind_a, ind_b))
      for ind in indexes:
        R[ind+(a,b)] = 1.0 / (ma*mb)

  return R

M2T = make_M2T().reshape(81,36)

def make_usym():
  """
    Make the tensor that goes from Mandel to tensors
  """
  R = np.zeros((3,3,6))
  for a in range(6):
    R[mandel[a][0],mandel[a][1],a] = 1.0/ mandel_mults[a]
    R[mandel[a][1],mandel[a][0],a] = 1.0/mandel_mults[a]
  
  return R

U = make_usym().reshape(9,6)

def make_sym():
  """
    Make the tensor that goes from tensors to Mandel
  """
  R = np.zeros((6,3,3))
  for a in range(6):
    R[a,mandel[a][0],mandel[a][1]] = mandel_mults[a]

  return R

S = make_sym().reshape(6,9)

def ms2ts(C):
  """
    Convert a Mandel notation stiffness matrix to a full stiffness tensor.
  """
  Ct = np.zeros((3,3,3,3))
  for a in range(6):
    for b in range(6):
      ind_a = itertools.permutations(mandel[a], r=2)
      ind_b = itertools.permutations(mandel[b], r=2)
      ma = mandel_mults[a]
      mb = mandel_mults[b]
      indexes = tuple(ai+bi for ai, bi in itertools.product(ind_a, ind_b))
      for ind in indexes:
        Ct[ind] = C[a,b] / (ma*mb)

  return Ct

def ms2ts_faster(C):
  """
    Get cute with symmetry
  """
  #pylint: disable=too-many-function-args
  return np.dot(M2T, C.flatten()).reshape(3,3,3,3)

def ts2ms(C):
  """
    Convert a stiffness tensor into a Mandel notation stiffness matrix
  """
  Cv = np.zeros((6,6))
  for i in range(6):
    for j in range(6):
      ma = mandel_mults[i]
      mb = mandel_mults[j]
      Cv[i,j] = C[mandel[i]+mandel[j]] * ma * mb

  return Cv

def sym(A):
  """
    Take a symmetric matrix to the Mandel convention vector.
  """
  return np.array([A[0,0], A[1,1], A[2,2], np.sqrt(2)*A[1,2], 
    np.sqrt(2)*A[0,2], np.sqrt(2)*A[0,1]])

def sym_faster(A):
  """
    Take a symmetric tensor to the Mandel convention with multiplication
  """
  return np.dot(S, A.flatten())

def usym(v):
  """
    Take a Mandel symmetric vector to the full matrix.
  """
  return np.array([
    [v[0], v[5]/np.sqrt(2), v[4]/np.sqrt(2)],
    [v[5]/np.sqrt(2), v[1], v[3]/np.sqrt(2)],
    [v[4]/np.sqrt(2), v[3]/np.sqrt(2), v[2]]
    ])

def usym_faster(v):
  """
    Take a tensor to Mandel using multiplication
  """
  return np.dot(U, v).reshape(3,3)



import unittest
import tempfile

import numpy as np
import scipy.interpolate as inter

from srlife import materials
from helpers import differentiate

class CommonThermalMaterial:
  def test_derivative(self):
    Tvals = np.linspace(np.min(self.T), np.max(self.T))[1:-1]

    for T in Tvals:
      self.assertTrue(np.isclose(self.obj.dconductivity(T),
        differentiate(lambda x: self.obj.conductivity(x), T)))
      self.assertTrue(np.isclose(self.obj.ddiffusivity(T),
        differentiate(lambda x: self.obj.diffusivity(x), T)))

class TestPiecewiseLinearThermalMaterial(CommonThermalMaterial, unittest.TestCase):
  def setUp(self):
    self.name = "Computonium"
    self.T = np.array([10.0,100.0,150.0,250.0])
    self.cond = np.array([25.0,50.0,10.0,30.0])
    self.diff = np.array([100.0,20.0,50.0,60.0])

    self.obj = materials.PiecewiseLinearThermalMaterial(self.name,
        self.T, self.cond, self.diff)

  def test_interpolate(self):
    T = 125.0
    self.assertTrue(np.isclose(inter.interp1d(self.T, self.cond)(T),
      self.obj.conductivity(T)))
    self.assertTrue(np.isclose(inter.interp1d(self.T, self.diff)(T),
      self.obj.diffusivity(T)))

  def test_store_recover(self):
    tf = tempfile.mktemp()

    self.obj.save(tf, "blah")

    rep = materials.ThermalMaterial.load(tf, "blah")

    self.assertEqual(rep.name, self.obj.name)
    self.assertTrue(np.allclose(rep.temps, self.obj.temps))
    self.assertTrue(np.allclose(rep.cond, self.obj.cond))
    self.assertTrue(np.allclose(rep.diff, self.obj.diff))


class CommonFluidMaterial:
  def test_derivtive(self):
    for T in self.Trange:
      self.assertAlmostEqual(self.obj.dcoefficient(self.mat, T),
          differentiate(lambda x: self.obj.coefficient(self.mat, x), T))

class TestConstantFluidMaterial(CommonFluidMaterial, unittest.TestCase):
  def setUp(self):
    self.value = 10.0
    self.data = {"Computonium": self.value}
    self.obj = materials.ConstantFluidMaterial(self.data)

    self.Trange = np.linspace(0, 100)
    self.mat = "Computonium"

  def test_recover(self):
    tfile = tempfile.mktemp()
    self.obj.save(tfile, "blah")

    comp = materials.FluidMaterial.load(tfile, "blah")

    self.assertAlmostEqual(self.obj.coefficient("Computonium", 0),
        comp.coefficient("Computonium", 0))

  def test_interpolate(self):
    self.assertAlmostEqual(self.obj.coefficient("Computonium", 0), self.value)


class TestPiecewiseLinearFluidMaterial(CommonFluidMaterial, unittest.TestCase):
  def setUp(self):
    self.data = {"Computonium": (np.array([10.0,100.0,200.0,300.0]), np.array([50,10.0,5.0,2.0])),
        "Other": (np.array([10,100,200.0,300.0]), np.array([10.0,15.0,20.0,25.0]))}
    self.obj = materials.PiecewiseLinearFluidMaterial(self.data)

    self.Trange = np.linspace(10,300)[1:-1]
    self.mat = "Other"

  def test_recover(self):
    tfile = tempfile.mktemp()
    self.obj.save(tfile, "blah")

    comp = materials.FluidMaterial.load(tfile, "blah")

    for k, (T, vals) in self.obj.data.items():
      self.assertTrue(k in comp.data)
      self.assertTrue(np.allclose(T, comp.data[k][0]))
      self.assertTrue(np.allclose(vals, comp.data[k][1]))

  def test_interpolate(self):
    T = 125.0
    mat = "Computonium"

    v1 = inter.interp1d(self.data[mat][0], self.data[mat][1])(T)

    self.assertAlmostEqual(v1, self.obj.coefficient(mat, T))

class TestStructuralMaterial(unittest.TestCase):
    def setUp(self):
        self.mat="Computonium"
        self.T = 850.15
        self.stress=100
        self.erange=5e-3
        self.true_inside_envelope=np.array([0.533, 0.2])
        self.false_inside_envelope=np.array([0.1, 0.77])
        self.error_inside_envelope=np.array([-0.01, 0.01])
        self.fatigue_pname="nFatigue"
        self.rupture_pname="avgRupture"
        self.cfinteraction_pname="cfinteraction"
        self.data={"nFatigue": {"curve1": {"T": "950","a": "-9.2e-01 -3.8e+00 -8.4e+00 -4.1e+00", "n": "3 2 1 0", "cutoff": "1.5e-3"},
        "curve2": {"T": "800", "a": "-19.2e-01 -3.8e+00 -18.4e+00 -14.1e+00", "n": "3 2 1 0", "cutoff": "1.5e-3"},
        "curve3": {"T": "500", "a": "-0.2e-01 -0.8e+00 -0.4e+00 -0.1e+00", "n": "3 2 1 0", "cutoff": "1.5e-3"}},
        "avgRupture": {"C": "17.16","a": "-1475.23 7289.41 -16642.64 35684.60", "n": "3 2 1 0"},
        "lbRupture": {"C": "10.0","a": "-1000.0 7000.0 -10000.0 30000.0", "n": "3 2 1 0"},
        "cfinteraction":"0.3 0.3"}

        self.structmat = materials.StructuralMaterial(self.data)

    def test_store_receover(self):
      tfile = tempfile.mktemp()
      self.structmat.save(tfile, "blah")
      test = materials.StructuralMaterial.load(tfile, "blah")

    def test_cycles_to_fail(self):
        fcycles = self.structmat.cycles_to_fail(self.fatigue_pname,self.T,self.erange)

        a=np.array([-9.2e-01, -3.8e+00, -8.4e+00, -4.1e+00])
        n=np.array([3,2,1,0])
        cutoff=1.5e-3
        sum = 0
        if self.erange<=cutoff:
            for (b,m) in zip(a,n):
                sum+=b*np.log10(cutoff)**m
        else:
            for (b,m) in zip(a,n):
                sum+=b*np.log10(self.erange)**m
        return 10**sum

        self.assertTrue(np.isclose(sum, fcycles))

    def test_rupturetime(self):
        rtime=self.structmat.time_to_rupture(self.rupture_pname, 
            np.array([self.T]), np.array([self.stress]))[0]

        C=17.16
        a=np.array([-1475.23, 7289.41, -16642.64, 35684.60])
        n=np.array([3,2,1,0])
        sum = 0
        for (b,m) in zip(a,n):
              sum+=b*np.log10(self.stress)**m
        sum=10**(sum/self.T-C)

        self.assertTrue(np.isclose(sum, rtime))

    def test_inside_envelope(self):
      """
        Test interaction_fatigue and interaction_creep
      """
      self.assertTrue(self.structmat.inside_envelope(self.cfinteraction_pname,self.true_inside_envelope[0],self.true_inside_envelope[1]))
      self.assertFalse(self.structmat.inside_envelope(self.cfinteraction_pname,self.false_inside_envelope[0],self.false_inside_envelope[1]))
      with self.assertRaises(ValueError):
          self.structmat.inside_envelope(self.cfinteraction_pname,self.error_inside_envelope[0],self.error_inside_envelope[1])



#pylint: disable=no-member

"""
  Complete receiver system structural solvers.  These operate
  under the assumption that you've already done the thermal problem.
"""

from abc import ABC, abstractmethod

import multiprocess

from srlife import spring, solverparams, receiver

class SystemSolver(ABC):
  """
    The system takes as input:
      1) A complete receiver, including temperatures
      2) The structural material

    It returns:
      1) The updated receiver, with all the structural results defined
  """
  @abstractmethod
  def solve(self, model, smat):
    """
      Solve the receiver system

      Parameters:
        model       receiver object
        smat        structural material
    """
    return

class SpringSystemSolver(SystemSolver):
  """
    Solve the receiver system using the network of springs model defined
    in the springs module
  """
  def __init__(self, pset = solverparams.ParameterSet(),
      rtol = 1.0e-6, atol = 1.0e-4, miter = 25, verbose = False):
    """
      Initialize the solver

      Additional parameters:
        pset        parameter dictionary overriding the default solver params
        rtol        relative tolerance
        atol        absolute tolerance
        miter       number of permissible nonlinear iterations
        verbose     print a lot of debug info
    """
    self.rtol = pset.get_default("rtol", rtol)
    self.atol = pset.get_default("atol", atol)
    self.miter = pset.get_default("miter", miter)
    self.verbose = pset.get_default("verbose", verbose)

  def solve(self, model, smat, ssolver, nthreads = 1,
      verbose = False, decorator = lambda fn, l: fn):
    """
      Solve a receiver model using a spring system

      Parameters:
        model           receiver model, with temperatures
        smat            structural material
        ssolver         structural solver to use

      Additional parameters:
        verbose         print stuff
        nthreads        number of threads to use
        decorator       progress decorator
    """
    network = self.make_network(model, smat, ssolver)
    subproblems = network.reduce_graph()

    # Simple heuristic for deciding who gets threads
    nprobs = len(subproblems)
    max_sub = max(sum(1 for i,j,data in sb.edges(data=True) 
      if isinstance(data['object'], spring.TubeSpring)) for sb in
      subproblems)

    if nprobs < max_sub:
      results = []
      if verbose:
        print("Solving substructures sequentially")
      for i,subproblem in enumerate(subproblems):
        if verbose:
          print("Solving substructure %i of %i" % (i+1, len(subproblems)))
        results.append(subproblem.solve_all(nthreads, decorator = decorator))
    else:
      sfn = lambda x: x.solve_all(1)
      if verbose: 
        print("Solving subproblems in parallel")
      with multiprocess.Pool(nthreads) as p:
        results = list(decorator(p.imap(sfn, subproblems), len(subproblems)))

    # Sigh, now need to copy the tube data into the correct location...
    for new, orig in zip(results, subproblems):
      for i,j,k in new.edges(keys=True):
        if isinstance(orig[i][j][k]['object'], spring.TubeSpring):
          orig[i][j][k]['object'].tube.copy_results(new[i][j][k]['object'].tube)

  def make_network(self, model, smat, ssolver):
    """
      Setup the complete spring network, given the receiver and
      material objects

      Parameters:
        model           fully-defined receiver object
        smat            structural material model
        ssolver         structural solver to use
    """
    network = spring.SpringNetwork(atol = self.atol, rtol = self.rtol,
        miter = self.miter, verbose = self.verbose)
    cn = 0
    network.add_node(cn)
    cn += 1
    for panel in model.panels.values():
      network.add_node(cn)
      cn += 1
      network.add_edge(0, cn-1, object = convert_to_spring(
        model.stiffness, smat, ssolver))
      top = cn-1
      for tube in panel.tubes.values():
        network.add_node(cn)
        cn += 1
        network.add_edge(top, cn-1, object = convert_to_spring(
          panel.stiffness, smat, ssolver))
        network.add_node(cn)
        cn += 1
        network.add_edge(cn-2, cn-1, object = convert_to_spring(tube,
          smat, ssolver))
        network.displacement_bc(cn-1, lambda t: 0.0)
    
    network.validate_setup()

    return network

def convert_to_spring(thing, smat, ssolver):
  """
    Translate an object into a spring

    Parameters:
      thing         object to convert
      smat          structural material
      ssolver       structural solver
  """
  if isinstance(thing, (float,int)):
    return spring.LinearSpring(thing)
  elif isinstance(thing, str):
    if thing not in ["disconnect", "rigid"]:
      raise ValueError("Special spring types are either 'disconnect' or 'rigid'!")
    return thing
  elif isinstance(thing, receiver.Tube):
    return spring.TubeSpring(thing, ssolver, smat)
  else:
    raise ValueError("Cannot convert object to spring!")



#!/usr/bin/env python3

import sys
sys.path.append('../../..')
from srlife import materials, thermal

from thermalsol import ManufacturedSolution

import numpy as np

problem = ManufacturedSolution("test", 1, lambda t, r: np.sin(r), 
    lambda t, k, alpha, r: k * np.sin(r) - k/r*np.cos(r))

def run_with(solver, material, fluid):
  """
    Run the standard problems with the provided solver and material
  """
  for problem in problems:
    res = problem.solve(solver, material, fluid)
    problem.plot_comparison(res)
    plt.show()

if __name__ == "__main__":
  solver = thermal.FiniteDifferenceImplicitThermalSolver()
  tmat = materials.ConstantThermalMaterial("Test", 10.0, 5.0)
  fmat = materials.ConstantFluidMaterial({"Test": 7.5})
  
  if len(sys.argv) == 2:
    n = int(sys.argv[1])
  else:
    n = 1

  for i in range(n):
    res = problem.solve(solver, tmat, fmat)
